# -*- coding: utf-8 -*-
"""
:mod:`orion.client` -- Python API
=================================

.. module:: client
   :platform: Unix
   :synopsis: Provides functions for communicating with `orion.core`.

"""
from orion.client.cli import (
    interrupt_trial, report_bad_trial, report_objective, report_results)
from orion.client.experiment import ExperimentClient
import orion.core.io.experiment_builder as experiment_builder
from orion.core.utils.exceptions import RaceCondition
from orion.core.utils.singleton import update_singletons
from orion.core.worker.producer import Producer
from orion.storage.base import setup_storage

__all__ = ['interrupt_trial', 'report_bad_trial', 'report_objective', 'report_results',
           'create_experiment', 'get_experiment', 'workon']


# pylint: disable=too-many-arguments
def create_experiment(
        name, version=None, space=None, algorithms=None,
        strategy=None, max_trials=None, storage=None, branching=None,
        max_idle_time=None, heartbeat=None, working_dir=None, debug=False):
    """Create an experiment

    There is 2 main scenarios

    1) The experiment is new

    ``name`` and ``space`` arguments are required, otherwise ``NoConfigurationError`` will be
    raised.

    All other arguments (``algorithms``, ``strategy``, ``max_trials``, ``storage``, ``branching``
    and ``working_dir``) will be replaced by system's defaults if ommited. The system's defaults can
    also be overriden in global configuration file as described for the database in
    :ref:`Database Configuration`. We do not recommand overriding the algorithm configuration using
    system's default, but overriding the storage configuration can be very convenient if the same
    storage is used for all your experiments.

    2) The experiment exist in the database.

    We can break down this scenario in two sub-scenarios for clarity.

    2.1) Only experiment name is given.

    The configuration will be fetched from database.

    2.2) Some other arguments than the name are given.

    The configuration will be fetched from database and given arguments will override them.
    ``max_trials`` may be overwritten in DB, but any other changes will lead to a branching. Instead
    of creating the experiment ``(name, version)``, it will create a new experiment
    ``(name, version+1)`` which will have the same configuration than ``(name, version)`` except for
    the differing arguments given by user. This new experiment will have access to trials of
    ``(name, version)``, adapted according to the differences between ``version`` and ``version+1``.
    A previous version can be accessed by specifying the ``version`` argument.

    Causes of experiment branching are:

    - Change of search space

        - New dimension

        - Different prior

        - Missing dimension

    - Change of algorithm

    - Change of strategy (Not implemented yet)

    - Change of code version (Only supported by commandline API for now)

    Parameters
    ----------
    name: str
        Name of the experiment
    version: int, optional
        Version of the experiment. Defaults to last existing version for a given ``name``
        or 1 for new experiment.
    space: dict, optional
        Optimization space of the algorithm. Should have the form ``dict(name='<prior>(args)')``.
    algorithms: str or dict, optional
        Algorithm used for optimization.
    strategy: str or dict, optional
        Parallel strategy to use to parallelize the algorithm.
    max_trials: int, optional
        Maximum number or trials before the experiment is considered done.
    storage: dict, optional
        Configuration of the storage backend.
    working_dir: str, optional
        Working directory created for the experiment inside which a unique folder will be created
        for each trial. Defaults to a temporary directory that is deleted at end of execution.
    max_idle_time: int, optional
        Maximum time the producer can spend trying to generate a new suggestion.
        Such timeout are generally caused by slow database, large number of
        concurrent workers leading to many race conditions or small search spaces
        with integer/categorical dimensions that may be fully explored.
        Defaults to `orion.core.config.worker.max_idle_time`.
    heartbeat: int, optional
        Frequency (seconds) at which the heartbeat of the trial is updated.
        If the heartbeat of a `reserved` trial is larger than twice the configured
        heartbeat, Oríon will reset the status of the trial to `interrupted`.
        This allows restoring lost trials (ex: due to killed worker).
        Defaults to `orion.core.config.worker.max_idle_time`.
    debug: bool, optional
        If using in debug mode, the storage config is overrided with legacy:EphemeralDB.
        Defaults to False.
    branching: dict, optional
        Arguments to control the branching.

        branch_to: str, optional
            Name of the experiment to branch to. The parent experiment will be the one specified by
            ``(name, version)``, and the child will be ``(branch_to, 1)``.
        branch_from: str, optional
            Name of the experiment to branch from.
            The parent experiment will be the one specified by
            ``(branch_from, last version)``, and the child will be ``(name, 1)``.
        manual_resolution: bool, optional
            Starts the prompt to resolve manually the conflicts. Defaults to False.
        algorithm_change: bool, optional
            Whether to automatically solve the algorithm conflict (change of algo config).
            Defaults to True.
        code_change_type: str, optional
            How to resolve code change automatically. Must be one of 'noeffect', 'unsure' or
            'break'.  Defaults to 'break'.
        cli_change_type: str, optional
            How to resolve cli change automatically. Must be one of 'noeffect', 'unsure' or 'break'.
            Defaults to 'break'.
        config_change_type: str, optional
            How to resolve config change automatically. Must be one of 'noeffect', 'unsure' or
            'break'.  Defaults to 'break'.

    Raises
    ------
    `orion.core.utils.SingletonAlreadyInstantiatedError`
        If the storage is already instantiated and given configuration is different.
        Storage is a singleton, you may only use one instance per process.
    `orion.core.utils.exceptions.NoConfigurationError`
        The experiment is not in database and no space is provided by the user.
    `orion.core.utils.exceptions.RaceCondition`
        There was a race condition during branching and new version cannot be infered because of
        that. Single race conditions are normally handled seemlessly. If this error gets raised, it
        means that different modifications occured during each race condition resolution. This is
        likely due to quick code change during experiment creation. Make sure your script is not
        generating files within your code repository.
    `orion.core.utils.exceptions.BranchingEvent`
        The configuration is different than the corresponding one in DB and the branching cannot be
        solved automatically. This usually happens if the version=x is specified but the experiment
        ``(name, x)`` already has a child ``(name, x+1)``. If you really need to branch from version
        ``x``, give it a new name to branch to with ``branching={'branch_to': <new_name>}``.
    `NotImplementedError`
        If the algorithm, storage or strategy specified is not properly installed.

    """
    setup_storage(storage=storage, debug=debug)

    try:
        experiment = experiment_builder.build(
            name, version=version, space=space, algorithms=algorithms,
            strategy=strategy, max_trials=max_trials, branching=branching,
            working_dir=working_dir)
    except RaceCondition:
        # Try again, but if it fails again, raise. Race conditions due to version increment should
        # only occur once in a short window of time unless code version is changing at a crazy pace.
        try:
            experiment = experiment_builder.build(
                name, version=version, space=space, algorithms=algorithms,
                strategy=strategy, max_trials=max_trials, branching=branching,
                working_dir=working_dir)
        except RaceCondition as e:
            raise RaceCondition(
                "There was a race condition during branching and new version cannot be infered "
                "because of that. Single race conditions are normally handled seemlessly. If this "
                "error gets raised, it means that different modifications occured during each race "
                "condition resolution. This is likely due to quick code change during experiment "
                "creation. Make sure your script is not generating files within your code "
                "repository.") from e

    producer = Producer(experiment, max_idle_time)

    return ExperimentClient(experiment, producer, heartbeat)


def get_experiment(name, version=None, storage=None):
    """
    Retrieve an existing experiment as :class:`orion.core.worker.experiment.ExperimentView`.

    Parameters
    ----------
    name: str
        The name of the experiment.
    version: int, optional
        Version to select. If None, last version will be selected. If version given is larger than
        largest version available, the largest version will be selected.
    storage: dict, optional
        Configuration of the storage backend.

    Returns
    -------
    An instance of :class:`orion.core.worker.experiment.ExperimentView` representing the experiment.

    Raises
    ------
    `orion.core.utils.exceptions.NoConfigurationError`
        The experiment is not in the database provided by the user.
    """
    setup_storage(storage)
    return experiment_builder.build_view(name, version)


def workon(function, space, name='loop', algorithms=None, max_trials=None):
    """Optimize a function over a given search space

    This will create a new experiment with an in-memory storage and optimize the given function
    until `max_trials` is reached or the `algorithm` is done
    (some algorithms like random search are never done).

    For informations on how to fetch results, see
    :py:class:`orion.client.experiment.ExperimentClient`.

    .. note::

        Each call to this function will create a separate in-memory storage.

    Parameters
    ----------
    name: str
        Name of the experiment
    version: int, optional
        Version of the experiment. Defaults to last existing version for a given `name`
        or 1 for new experiment.
    space: dict, optional
        Optimization space of the algorithm. Should have the form `dict(name='<prior>(args)')`.
    algorithms: str or dict, optional
        Algorithm used for optimization.
    max_trials: int, optional
        Maximum number or trials before the experiment is considered done.

    Raises
    ------
    `NotImplementedError`
        If the algorithm specified is not properly installed.

    """
    # Clear singletons and keep pointers to restore them.
    singletons = update_singletons()

    try:
        setup_storage(storage={'type': 'legacy', 'database': {'type': 'EphemeralDB'}})

        experiment = experiment_builder.build(
            name, version=1, space=space, algorithms=algorithms,
            strategy='NoParallelStrategy', max_trials=max_trials)

        producer = Producer(experiment)

        experiment_client = ExperimentClient(experiment, producer)
        experiment_client.workon(function, max_trials=max_trials)
    finally:
        # Restore singletons
        update_singletons(singletons)

    return experiment_client



# -*- coding: utf-8 -*-
"""
:mod:`orion.algo.tpe` -- Tree-structured Parzen Estimator Approach
==================================================================

.. module:: tpe
   :platform: Unix
   :synopsis: Tree-structured Parzen Estimator Approach

"""
import logging

import numpy
from scipy.stats import norm

from orion.algo.base import BaseAlgorithm
from orion.core.utils.points import flatten_dims, regroup_dims

logger = logging.getLogger(__name__)


def compute_max_ei_point(points, below_likelis, above_likelis):
    """Compute ei among points based on their log likelihood and return the point with max ei.

    :param points: list of point with real values.
    :param below_likelis: list of log likelihood for each point in the good GMM.
    :param above_likelis: list of log likelihood for each point in the bad GMM.
    """
    max_ei = -numpy.inf
    point_index = 0
    for i, (lik_b, lik_a) in enumerate(zip(below_likelis, above_likelis)):
        ei = lik_b - lik_a
        if ei > max_ei:
            max_ei = ei
            point_index = i
    return points[point_index]


def ramp_up_weights(total_num, flat_num, equal_weight):
    """Adjust weights of observed trials.

    :param total_num: total number of observed trials.
    :param flat_num: the number of the most recent trials which
        get the full weight where the others will be applied with a linear ramp
        from 0 to 1.0. It will only take effect if equal_weight is False.
    :param equal_weight: whether all the observed trails share the same weights.
    """
    if total_num < flat_num or equal_weight:
        return numpy.ones(total_num)

    ramp_weights = numpy.linspace(1.0 / total_num, 1.0, num=total_num - flat_num)
    flat_weights = numpy.ones(flat_num)
    return numpy.concatenate([ramp_weights, flat_weights])


# pylint:disable=assignment-from-no-return
def adaptive_parzen_estimator(mus, low, high,
                              prior_weight=1.0,
                              equal_weight=False,
                              flat_num=25):
    """Return the sorted mus, the corresponding sigmas and weights with adaptive kernel estimator.

    This adaptive parzen window estimator is based on the original papers and also refer the use of
    prior mean in `this implementation
    <https://github.com/hyperopt/hyperopt/blob/40bd617a0ca47e09368fb65919464ba0bfa85962/hyperopt/tpe.py#L400-L468>`_.

    :param mus: list of real values for observed mus.
    :param low: real value for lower bound of points.
    :param high: real value for upper bound of points.
    :param prior_weight: real value for the weight of the prior mean.
    :param equal_weight: bool value indicating if all points with equal weights.
    :param flat_num: int value indicating the number of the most recent trials which
        get the full weight where the others will be applied with a linear ramp
        from 0 to 1.0. It will only take effect if equal_weight is False.
    """
    mus = numpy.asarray(mus)

    prior_mu = (low + high) * 0.5
    prior_sigma = (high - low) * 1.0

    size = len(mus)
    if size > 1:
        order = numpy.argsort(mus)
        sorted_mus = mus[order]
        prior_mu_pos = numpy.searchsorted(sorted_mus, prior_mu)

        weights = ramp_up_weights(size, flat_num, equal_weight)

        mixture_mus = numpy.zeros(size + 1)
        mixture_mus[:prior_mu_pos] = sorted_mus[:prior_mu_pos]
        mixture_mus[prior_mu_pos] = prior_mu
        mixture_mus[prior_mu_pos + 1:] = sorted_mus[prior_mu_pos:]

        mixture_weights = numpy.ones(size + 1)
        mixture_weights[:prior_mu_pos] = weights[:prior_mu_pos]
        mixture_weights[prior_mu_pos] = prior_weight
        mixture_weights[prior_mu_pos + 1:] = weights[prior_mu_pos:]

        sigmas = numpy.ones(size + 1)
        sigmas[0] = mixture_mus[1] - mixture_mus[0]
        sigmas[-1] = mixture_mus[-1] - mixture_mus[-2]
        sigmas[1:-1] = numpy.maximum((mixture_mus[1:-1] - mixture_mus[0:-2]),
                                     (mixture_mus[2:] - mixture_mus[1:-1]))
        sigmas = numpy.clip(sigmas, prior_sigma / max(10, numpy.sqrt(size)), prior_sigma)

    else:
        if prior_mu < mus[0]:

            mixture_mus = numpy.array([prior_mu, mus[0]])
            sigmas = numpy.array([prior_sigma, prior_sigma * 0.5])
            mixture_weights = numpy.array([prior_weight, 1.0])
        else:
            mixture_mus = numpy.array([mus[0], prior_mu])
            sigmas = numpy.array([prior_sigma * 0.5, prior_sigma])
            mixture_weights = numpy.array([1.0, prior_weight])

    weights = mixture_weights / mixture_weights.sum()

    return mixture_mus, sigmas, weights


class TPE(BaseAlgorithm):
    """Tree-structured Parzen Estimator (TPE) algorithm is one of Sequential Model-Based
    Global Optimization (SMBO) algorithms, which will build models to propose new points based
    on the historical observed trials.

    Instead of modeling p(y|x) like other SMBO algorithms, TPE models p(x|y) and p(y),
    and p(x|y) is modeled by transforming that generative process, replacing the distributions of
    the configuration prior with non-parametric densities.

    The TPE defines p(x|y) using two such densities l(x) and g(x) while l(x) is distribution of
    good points and g(x) is the distribution of bad points. New point candidates will be sampled
    with l(x) and Expected Improvement (EI) optimization scheme will be used to find the most
    promising point among the candidates.

    For more information on the algorithm, see original papers at:

    - `Algorithms for Hyper-Parameter Optimization
      <https://papers.nips.cc/paper/4443-algorithms-for-hyper-parameter-optimization.pdf>`_
    - `Making a Science of Model Search: Hyperparameter Optimizationin Hundreds of Dimensions
      for Vision Architectures <http://proceedings.mlr.press/v28/bergstra13.pdf>`_

    Parameters
    ----------
    space: `orion.algo.space.Space`
        Optimisation space with priors for each dimension.
    seed: None, int or sequence of int
        Seed to sample initial points and candidates points.
        Default: ``None``
    n_initial_points: int
        Number of initial points randomly sampled.
        Default: ``20``
    n_ei_candidates: int
        Number of candidates points sampled for ei compute.
        Default: ``24``
    gamma: real
        Ratio to split the observed trials into good and bad distributions.
        Default: ``0.25``
    equal_weight: bool
        True to set equal weights for observed points.
        Default: ``False``
    prior_weight: int
        The weight given to the prior point of the input space.
        Default: ``1.0``
    full_weight_num: int
        The number of the most recent trials which get the full weight where the others will be
        applied with a linear ramp from 0 to 1.0. It will only take effect if equal_weight
        is False.

    """

    # pylint:disable=too-many-arguments
    def __init__(self, space, seed=None,
                 n_initial_points=20, n_ei_candidates=24,
                 gamma=0.25, equal_weight=False,
                 prior_weight=1.0, full_weight_num=25):

        super(TPE, self).__init__(space,
                                  seed=seed,
                                  n_initial_points=n_initial_points,
                                  n_ei_candidates=n_ei_candidates,
                                  gamma=gamma,
                                  equal_weight=equal_weight,
                                  prior_weight=prior_weight,
                                  full_weight_num=full_weight_num)

        for dimension in self.space.values():

            if dimension.type != 'fidelity' and \
                    dimension.prior_name not in ['uniform', 'reciprocal', 'int_uniform', 'choices']:
                raise ValueError("TPE now only supports uniform, loguniform, uniform discrete "
                                 "and choices as prior.")

            shape = dimension.shape
            if shape and len(shape) != 1:
                raise ValueError("TPE now only supports 1D shape.")

        if n_initial_points < 2:
            n_initial_points = 2
            logger.warning('n_initial_points %s is not valid, set n_initial_points = 2',
                           str(n_initial_points))

        if n_ei_candidates < 1:
            n_ei_candidates = 1
            logger.warning('n_ei_candidates %s is not valid, set n_ei_candidates = 1',
                           str(n_ei_candidates))

        self.seed_rng(seed)

    def seed_rng(self, seed):
        """Seed the state of the random number generator.

        :param seed: Integer seed for the random number generator.
        """
        self.rng = numpy.random.RandomState(seed)

    @property
    def state_dict(self):
        """Return a state dict that can be used to reset the state of the algorithm."""
        _state_dict = super(TPE, self).state_dict

        _state_dict['rng_state'] = self.rng.get_state()
        _state_dict['seed'] = self.seed
        return _state_dict

    def set_state(self, state_dict):
        """Reset the state of the algorithm based on the given state_dict

        :param state_dict: Dictionary representing state of an algorithm
        """
        super(TPE, self).set_state(state_dict)

        self.seed_rng(state_dict['seed'])
        self.rng.set_state(state_dict['rng_state'])

    def suggest(self, num=1):
        """Suggest a `num` of new sets of parameters. Randomly draw samples
        from the import space and return them.

        :param num: how many sets to be suggested.

        .. note:: New parameters must be compliant with the problem's domain
           `orion.algo.space.Space`.
        """
        if num > 1:
            raise ValueError("TPE should suggest only one point.")

        samples = []
        if len(self._trials_info) < self.n_initial_points:
            new_point = self.space.sample(1, seed=tuple(self.rng.randint(0, 1000000, size=3)))[0]
            samples.append(new_point)
        else:
            point = []
            below_points, above_points = self.split_trials()

            below_points = [flatten_dims(point, self.space) for point in below_points]
            above_points = [flatten_dims(point, self.space) for point in above_points]
            below_points = list(map(list, zip(*below_points)))
            above_points = list(map(list, zip(*above_points)))

            idx = 0
            for dimension in self.space.values():

                shape = dimension.shape
                if not shape:
                    shape = (1,)

                if dimension.type == 'real':
                    points = self._sample_real_dimension(dimension, shape[0],
                                                         below_points[idx: idx + shape[0]],
                                                         above_points[idx: idx + shape[0]])
                elif dimension.type == 'integer' and dimension.prior_name == 'int_uniform':
                    points = self.sample_one_dimension(dimension, shape[0],
                                                       below_points[idx: idx + shape[0]],
                                                       above_points[idx: idx + shape[0]],
                                                       self._sample_int_point)
                elif dimension.type == 'categorical' and dimension.prior_name == 'choices':
                    points = self.sample_one_dimension(dimension, shape[0],
                                                       below_points[idx: idx + shape[0]],
                                                       above_points[idx: idx + shape[0]],
                                                       self._sample_categorical_point)
                elif dimension.type == 'fidelity':
                    # fidelity dimension
                    points = dimension.sample(num)
                else:
                    raise NotImplementedError()

                if len(points) < shape[0]:
                    logger.warning('TPE failed to sample new point with configuration %s',
                                   self.configuration)
                    return None

                idx += shape[0]
                point += points

            point = regroup_dims(point, self.space)
            samples.append(point)

        return samples

    # pylint:disable=no-self-use
    def sample_one_dimension(self, dimension, shape_size, below_points, above_points, sampler):
        """Sample values for a dimension

        :param dimension: Dimension.
        :param shape_size: 1D Shape Size of the Real Dimension.
        :param below_points: good points with shape (m, n), m=shape_size.
        :param above_points: bad points with shape (m, n), m=shape_size.
        :param sampler: method to sample one value for upon the dimension.
        """
        points = []

        for j in range(shape_size):
            new_point = sampler(dimension, below_points[j], above_points[j])
            if new_point is not None:
                points.append(new_point)

        return points

    def _sample_real_dimension(self, dimension, shape_size, below_points, above_points):
        """Sample values for real dimension"""
        if dimension.prior_name == 'uniform':
            return self.sample_one_dimension(dimension, shape_size, below_points, above_points,
                                             self._sample_real_point)
        elif dimension.prior_name == 'reciprocal':
            return self.sample_one_dimension(dimension, shape_size, below_points, above_points,
                                             self._sample_loguniform_real_point)
        else:
            raise NotImplementedError()

    def _sample_loguniform_real_point(self, dimension, below_points, above_points):
        """Sample one value for real dimension in a loguniform way"""
        return self._sample_real_point(dimension, below_points, above_points, is_log=True)

    def _sample_real_point(self, dimension, below_points, above_points, is_log=False):
        """Sample one value for real dimension based on the observed good and bad points"""
        low, high = dimension.interval()
        if is_log:
            low = numpy.log(low)
            high = numpy.log(high)
            below_points = numpy.log(below_points)
            above_points = numpy.log(above_points)

        below_mus, below_sigmas, below_weights = \
            adaptive_parzen_estimator(below_points, low, high, self.prior_weight,
                                      self.equal_weight, flat_num=self.full_weight_num)
        above_mus, above_sigmas, above_weights = \
            adaptive_parzen_estimator(above_points, low, high, self.prior_weight,
                                      self.equal_weight, flat_num=self.full_weight_num)

        gmm_sampler_below = GMMSampler(self, below_mus, below_sigmas,
                                       low, high, below_weights)
        gmm_sampler_above = GMMSampler(self, above_mus, above_sigmas,
                                       low, high, above_weights)

        candidate_points = gmm_sampler_below.sample(self.n_ei_candidates)
        if candidate_points:
            lik_blow = gmm_sampler_below.get_loglikelis(candidate_points)
            lik_above = gmm_sampler_above.get_loglikelis(candidate_points)
            new_point = compute_max_ei_point(candidate_points, lik_blow, lik_above)

            if is_log:
                new_point = numpy.exp(new_point)

            return new_point

        return None

    def _sample_int_point(self, dimension, below_points, above_points):
        """Sample one value for integer dimension based on the observed good and bad points"""
        low, high = dimension.interval()
        choices = range(low, high)

        below_points = numpy.array(below_points).astype(int) - low
        above_points = numpy.array(above_points).astype(int) - low

        sampler_below = CategoricalSampler(self, below_points, choices)
        candidate_points = sampler_below.sample(self.n_ei_candidates)

        if list(candidate_points):
            sampler_above = CategoricalSampler(self, above_points, choices)

            lik_below = sampler_below.get_loglikelis(candidate_points)
            lik_above = sampler_above.get_loglikelis(candidate_points)

            new_point = compute_max_ei_point(candidate_points, lik_below, lik_above)
            new_point = new_point + low
            return new_point

        return None

    def _sample_categorical_point(self, dimension, below_points, above_points):
        """Sample one value for categorical dimension based on the observed good and bad points"""
        choices = dimension.interval()

        below_points = [choices.index(point) for point in below_points]
        above_points = [choices.index(point) for point in above_points]

        sampler_below = CategoricalSampler(self, below_points, choices)
        candidate_points = sampler_below.sample(self.n_ei_candidates)

        if list(candidate_points):
            sampler_above = CategoricalSampler(self, above_points, choices)

            lik_below = sampler_below.get_loglikelis(candidate_points)
            lik_above = sampler_above.get_loglikelis(candidate_points)

            new_point_index = compute_max_ei_point(candidate_points, lik_below, lik_above)
            new_point = choices[new_point_index]

            return new_point
        return None

    def split_trials(self):
        """Split the observed trials into good and bad ones based on the ratio `gamma``"""
        sorted_trials = sorted(self._trials_info.values(), key=lambda x: x[1]['objective'])
        sorted_points = [list(points) for points, results in sorted_trials]

        split_index = int(numpy.ceil(self.gamma * len(sorted_points)))

        below = sorted_points[:split_index]
        above = sorted_points[split_index:]

        return below, above

    def observe(self, points, results):
        """Observe evaluation `results` corresponding to list of `points` in
        space.

        A simple random sampler though does not take anything into account.
        """
        super(TPE, self).observe(points, results)


class GMMSampler():
    """Gaussian Mixture Model Sampler for TPE algorithm

    Parameters
    ----------
    tpe: `TPE` algorithm
        The tpe algorithm object which this sampler will be part of.
    mus: list
        mus for each Gaussian components in the GMM.
        Default: ``None``
    sigmas: list
        sigmas for each Gaussian components in the GMM.
    low: real
        Lower bound of the sampled points.
    high: real
        Upper bound of the sampled points.
    weights: list
        Weights for each Gaussian components in the GMM
        Default: ``None``

    """

    def __init__(self, tpe, mus, sigmas, low, high, weights=None):
        self.tpe = tpe

        self.mus = mus
        self.sigmas = sigmas
        self.low = low
        self.high = high
        self.weights = weights if weights is not None else len(mus) * [1.0 / len(mus)]

        self.pdfs = []
        self._build_mixture()

    def _build_mixture(self):
        """Build the Gaussian components in the GMM"""
        for mu, sigma in zip(self.mus, self.sigmas):
            self.pdfs.append(norm(mu, sigma))

    def sample(self, num=1):
        """Sample required number of points"""
        point = []
        for _ in range(num):
            pdf = numpy.argmax(self.tpe.rng.multinomial(1, self.weights))
            new_points = self.pdfs[pdf].rvs(size=5)
            for pt in new_points:
                if self.low <= pt < self.high:
                    point.append(pt)
                    break

        return point

    def get_loglikelis(self, points):
        """Return the log likelihood for the points"""
        points = numpy.array(points)
        weight_likelis = [numpy.log(self.weights[i] * pdf.pdf(points))
                          for i, pdf in enumerate(self.pdfs)]
        weight_likelis = numpy.array(weight_likelis)
        # (num_weights, num_points) => (num_points, num_weights)
        weight_likelis = weight_likelis.transpose()

        # log-sum-exp trick
        max_likeli = numpy.nanmax(weight_likelis, axis=1)
        point_likeli = max_likeli + numpy.log(numpy.nansum
                                              (numpy.exp(weight_likelis - max_likeli[:, None]),
                                               axis=1))

        return point_likeli


class CategoricalSampler():
    """Categorical Sampler for discrete integer and categorical choices

    Parameters
    ----------
    tpe: `TPE` algorithm
        The tpe algorithm object which this sampler will be part of.
    observations: list
        Observed values in the dimension
    choices: list
        Candidate values for the dimension

    """

    def __init__(self, tpe, observations, choices):
        self.tpe = tpe
        self.obs = observations
        self.choices = choices

        self._build_multinomial_weights()

    def _build_multinomial_weights(self):
        """Build weights for categorical distribution based on observations"""
        weights_obs = ramp_up_weights(len(self.obs),
                                      self.tpe.full_weight_num, self.tpe.equal_weight)
        counts_obs = numpy.bincount(self.obs, minlength=len(self.choices), weights=weights_obs)
        counts_obs = counts_obs + self.tpe.prior_weight
        self.weights = counts_obs / counts_obs.sum()

    def sample(self, num=1):
        """Sample required number of points"""
        samples = self.tpe.rng.multinomial(n=1, pvals=self.weights, size=num)

        assert samples.shape == (num,) + (len(self.weights),)

        samples_index = samples.argmax(-1)
        assert samples_index.shape == (num,)

        return samples_index

    def get_loglikelis(self, points):
        """Return the log likelihood for the points"""
        return numpy.log(numpy.asarray(self.weights)[points])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.utils.working_dir`."""
import os
import shutil

import pytest

from orion.core.utils.working_dir import WorkingDir


@pytest.fixture
def path(tmp_path):
    """Return a path as a string."""
    return str(tmp_path) + "/hi_hello"


def test_create_permanent_dir(tmp_path, path):
    """Check if a permanent directory is created."""
    with WorkingDir(tmp_path, temp=False, prefix="hi", suffix="_hello"):
        assert os.path.exists(path)

    assert os.path.exists(path)


def test_temp_dir_when_exists(tmp_path, path):
    """Check if a permanent directory is deleted."""
    os.mkdir(path)

    with WorkingDir(tmp_path, temp=True, prefix="hi", suffix="_hello"):
        assert os.path.exists(path)

    assert os.path.exists(path)

    shutil.rmtree(path)


def test_create_temp_dir(tmp_path):
    """Check if a temporary directory is created."""
    with WorkingDir(tmp_path, prefix="hi", suffix="_hello") as w:
        assert os.path.exists(w)

    assert not os.path.exists(w)



# -*- coding: utf-8 -*-
"""
:mod:`orion.analysis` -- Provides HPO analysis tools
====================================================

.. module:: analysis
   :platform: Unix
   :synopsis: Provides agnostic HPO analysis tools
"""

from orion.analysis.regret import regret

__all__ = ['regret']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.cli.status` -- Module to status experiments
============================================================

.. module:: status
   :platform: Unix
   :synopsis: List the trials and their statuses for experiments.

"""
import collections
import logging

import tabulate

from orion.core.cli import base as cli
import orion.core.io.experiment_builder as experiment_builder
from orion.storage.base import get_storage

log = logging.getLogger(__name__)
SHORT_DESCRIPTION = "Gives an overview of experiments' trials"
DESCRIPTION = """
This command outputs the status of the different trials inside every experiment or a
specific EVC tree. It can either give you an overview of the different trials status, i.e.,
the number of currently completed trials and so on, or, it can give you a deeper view of the
experiments by outlining every single trial, its status and its objective.
"""


def add_subparser(parser):
    """Add the subparser that needs to be used for this command"""
    status_parser = parser.add_parser('status', help=SHORT_DESCRIPTION, description=DESCRIPTION)

    cli.get_basic_args_group(status_parser)

    status_parser.add_argument(
        '-a', '--all', action="store_true",
        help="Show all trials line by line. Otherwise, they are all aggregated by status")

    status_parser.add_argument(
        '-C', '--collapse', action="store_true",
        help=("Aggregate together results of all child experiments. Otherwise they are all "
              "printed hierarchically"))

    status_parser.add_argument(
        '-e', '--expand-versions', action='store_true',
        help=("Show all the version of every experiments instead of only the latest one")
        )

    status_parser.set_defaults(func=main)

    return status_parser


def main(args):
    """Fetch config and status experiments"""
    config = experiment_builder.get_cmd_config(args)
    experiment_builder.setup_storage(config.get('storage'))

    args['all_trials'] = args.pop('all', False)

    experiments = get_experiments(args)

    if not experiments:
        print("No experiment found")
        return

    if args.get('name'):
        print_evc([experiments[0]], **args)
        return

    if args.get('version'):
        if args.get('collapse') or args.get('expand_versions'):
            raise RuntimeError("Cannot fetch specific version of experiments with --collapse "
                               "or --expand-versions.")

    print_evc(experiments, **args)


# pylint: disable=unused-argument
def print_evc(experiments, version=None, all_trials=False, collapse=False,
              expand_versions=False, **kwargs):
    """Print each EVC tree

    Parameters
    ----------
    args: dict
        Commandline arguments.

    """
    for exp in experiments:
        experiment = experiment_builder.build_view(exp.name, version)
        if version is None:
            expand_experiment = exp
        else:
            expand_experiment = experiment
        expand = expand_versions or _has_named_children(expand_experiment)
        if expand and not collapse:
            print_status_recursively(expand_experiment, all_trials=all_trials)
        else:
            print_status(experiment, all_trials=all_trials, collapse=True)


def get_experiments(args):
    """Return the different experiments.

    Parameters
    ----------
    args: dict
        Commandline arguments.

    """
    projection = {'name': 1, 'version': 1, 'refers': 1}

    query = {'name': args['name']} if args.get('name') else {}
    experiments = get_storage().fetch_experiments(query, projection)

    if args['name']:
        root_experiments = experiments
    else:
        root_experiments = [exp for exp in experiments
                            if exp['refers'].get('root_id', exp['_id']) == exp['_id']]

    return [experiment_builder.build_view(name=exp['name'], version=exp.get('version', 1))
            for exp in root_experiments]


def _has_named_children(exp):
    return any(node.name != exp.name for node in exp.node)


def print_status_recursively(exp, depth=0, **kwargs):
    """Print the status recursively of the children of the current experiment.

    Parameters
    ----------
    exp: `orion.core.worker.Experiment`
        The current experiment to print.
    depth: int
        The current depth of the tree.

    """
    print_status(exp, offset=depth * 2, **kwargs)

    for child in exp.node.children:
        print_status_recursively(child.item, depth + 1, **kwargs)


def print_status(exp, offset=0, all_trials=False, collapse=False):
    """Print the status of the current experiment.

    Parameters
    ----------
    offset: int, optional
        The number of tabs to the right this experiment is.
    all_trials: bool, optional
        Print all trials individually
    collapse: bool, optional
        Fetch trials for entire EVCTree. Defaults to False.

    """
    trials = exp.fetch_trials(with_evc_tree=collapse)

    exp_title = exp.node.tree_name
    print(" " * offset, exp_title, sep="")
    print(" " * offset, "=" * len(exp_title), sep="")

    if all_trials:
        print_all_trials(trials, offset=offset)
    else:
        print_summary(trials, offset=offset)


def print_summary(trials, offset=0):
    """Print a summary of the current experiment.

    Parameters
    ----------
    trials: list
        Trials to summarize.
    offset: int, optional
        The number of tabs to the right this experiment is.

    """
    status_dict = collections.defaultdict(list)

    for trial in trials:
        status_dict[trial.status].append(trial)

    headers = ['status', 'quantity']

    lines = []
    for status, c_trials in sorted(status_dict.items()):
        line = [status, len(c_trials)]

        if c_trials[0].objective:
            headers.append('min {}'.format(c_trials[0].objective.name))
            line.append(min(trial.objective.value for trial in c_trials if trial.objective))

        lines.append(line)

    if trials:
        grid = tabulate.tabulate(lines, headers=headers)
        tab = " " * offset
        print(tab + ("\n" + tab).join(grid.split("\n")))
    else:
        print(" " * offset, 'empty', sep="")

    print("\n")


def print_all_trials(trials, offset=0):
    """Print all trials of the current experiment individually.

    Parameters
    ----------
    trials: list
        Trials to list in terminal.
    offset: int, optional
        The number of tabs to the right this experiment is.

    """
    headers = ['id', 'status', 'best objective']
    lines = []

    for trial in sorted(trials, key=lambda t: t.status):
        line = [trial.id, trial.status]

        if trial.objective:
            headers[-1] = 'min {}'.format(trial.objective.name)
            line.append(trial.objective.value)

        lines.append(line)

    if not trials:
        lines.append(['empty', '', ''])

    grid = tabulate.tabulate(lines, headers=headers)
    tab = " " * offset
    print(tab + ("\n" + tab).join(grid.split("\n")))

    print("\n")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.io.interactive_commands.branching_prompt`."""
import shlex

import pytest

from orion.core import evc
from orion.core.io.experiment_branch_builder import ExperimentBranchBuilder
from orion.core.io.interactive_commands.branching_prompt import BranchingPrompt
from orion.core.io.space_builder import DimensionBuilder


@pytest.fixture
def new_cat_dimension_conflict(old_config, new_config):
    """Generate a new dimension conflict with categorical prior for new experiment configuration"""
    name = 'new-cat'
    prior = 'choices(["hello", 2])'
    dimension = DimensionBuilder().build(name, prior)
    return evc.conflicts.NewDimensionConflict(old_config, new_config, dimension, prior)


@pytest.fixture
def missing_conflict_with_identical_prior(old_config, new_config, new_dimension_conflict):
    """Generate a missing dimension conflict which have the same prior as the new dim conflict"""
    name = 'missing-idem'
    prior = new_dimension_conflict.prior
    dimension = DimensionBuilder().build(name, prior)
    return evc.conflicts.MissingDimensionConflict(old_config, new_config, dimension, prior)


@pytest.fixture
def missing_cat_dimension_conflict(old_config, new_config):
    """Generate a missing dimension conflict with categorical prior for new experiment
    configuration
    """
    name = 'missing-cat'
    prior = 'choices(["goodbye", 5])'
    dimension = DimensionBuilder().build(name, prior)
    return evc.conflicts.MissingDimensionConflict(old_config, new_config, dimension, prior)


@pytest.fixture
def conflicts(new_dimension_conflict, new_cat_dimension_conflict,
              changed_dimension_conflict,
              missing_dimension_conflict, missing_cat_dimension_conflict,
              missing_conflict_with_identical_prior,
              algorithm_conflict, code_conflict, cli_conflict, config_conflict,
              experiment_name_conflict):
    """Create a container for conflicts with one of each types for testing purposes"""
    conflicts = evc.conflicts.Conflicts()
    conflicts.register(new_dimension_conflict)
    conflicts.register(new_cat_dimension_conflict)
    conflicts.register(changed_dimension_conflict)
    conflicts.register(missing_dimension_conflict)
    conflicts.register(missing_cat_dimension_conflict)
    conflicts.register(missing_conflict_with_identical_prior)
    conflicts.register(algorithm_conflict)
    conflicts.register(code_conflict)
    conflicts.register(config_conflict)
    conflicts.register(cli_conflict)
    conflicts.register(experiment_name_conflict)
    return conflicts


@pytest.fixture
def branch_builder(conflicts):
    """Generate the experiment branch builder"""
    return ExperimentBranchBuilder(conflicts, {'manual_resolution': True})


@pytest.fixture
def branch_solver_prompt(branch_builder):
    """Generate the branch solved prompt"""
    return BranchingPrompt(branch_builder)


class TestCommands(object):
    """Test the commands of the prompt"""

    def test_add_dim(self, conflicts, branch_solver_prompt):
        """Verify that dimension is added properly"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_add("new")
        assert len(conflicts.get_resolved()) == 1

    def test_add_bad_dim(self, capsys, conflicts, branch_solver_prompt):
        """Verify that error message is given when dimension does not exist"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_add("idontexist")
        out, err = capsys.readouterr()
        assert "Dimension name 'idontexist' not found in conflicts" in out
        assert len(conflicts.get_resolved()) == 0

    def test_add_dim_twice(self, capsys, conflicts, branch_solver_prompt):
        """Verify that error message is given trying to solve twice the same conflict"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_add("new")
        assert len(conflicts.get_resolved()) == 1
        branch_solver_prompt.do_add("new")
        out, err = capsys.readouterr()
        assert "Dimension name \'new\' not found in conflicts" in out
        assert len(conflicts.get_resolved()) == 1

    def test_add_dim_with_default(self, conflicts, branch_solver_prompt):
        """Verify that dimension is added with given default value"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_add("new --default-value=9")
        assert len(conflicts.get_resolved()) == 1
        assert conflicts.get_resolved()[0].resolution.default_value == 9

    def test_add_dim_with_bad_default(self, capsys, conflicts, branch_solver_prompt):
        """Verify that error message is given for bad default value"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_add("new --default-value='bad'")
        out, err = capsys.readouterr()
        assert "could not convert string to float: 'bad'" in out
        assert len(conflicts.get_resolved()) == 0

    def test_add_dim_with_cat(self, conflicts, branch_solver_prompt):
        """Verify that categorical dimensions is added properly"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_add("new_cat")
        assert len(conflicts.get_resolved()) == 1

    def test_add_dim_with_cat_default(self, conflicts, branch_solver_prompt):
        """Verify that string and non-string default value are properly parsed"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_add("new_cat --default-value='hello'")
        assert len(conflicts.get_resolved()) == 1
        assert conflicts.get_resolved()[0].resolution.default_value == 'hello'

        conflicts.revert(conflicts.get_resolved()[0].resolution)
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_add("new_cat --default-value=2")
        assert len(conflicts.get_resolved()) == 1
        assert conflicts.get_resolved()[0].resolution.default_value == 2

    def test_add_dim_with_cat_bad_default(self, capsys, conflicts, branch_solver_prompt):
        """Verify that error message is given for default value of invalid category"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_add("new_cat --default-value='bad'")
        out, err = capsys.readouterr()
        assert "Invalid category: bad" in out
        assert len(conflicts.get_resolved()) == 0

    def test_reset_add(self, conflicts, branch_solver_prompt):
        """Verify that addition resolution is reverted"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_add("new")
        assert len(conflicts.get_resolved()) == 1
        branch_solver_prompt.do_reset("'{}'".format(str(conflicts.get_resolved()[0].resolution)))
        assert len(conflicts.get_resolved()) == 0

    def test_change_dim(self, conflicts, branch_solver_prompt):
        """Verify that changed resolution is created"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_add("changed")
        assert len(conflicts.get_resolved()) == 1

    def test_change_dim_twice(self, capsys, conflicts, branch_solver_prompt):
        """Verify that error message is given trying to solve twice the same conflict"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_add("changed")
        assert len(conflicts.get_resolved()) == 1
        branch_solver_prompt.do_add("changed")
        out, err = capsys.readouterr()
        assert "Dimension name \'changed\' not found in conflicts" in out
        assert len(conflicts.get_resolved()) == 1

    def test_reset_change(self, conflicts, branch_solver_prompt):
        """Verify that changed resolution is reverted"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_add("changed")
        assert len(conflicts.get_resolved()) == 1
        branch_solver_prompt.do_reset("'{}'".format(str(conflicts.get_resolved()[0].resolution)))
        assert len(conflicts.get_resolved()) == 0

    def test_remove_dim(self, conflicts, branch_solver_prompt):
        """Verify that missing dimension is removed"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_remove("missing")
        assert len(conflicts.get_resolved()) == 1

    def test_remove_bad_dim(self, capsys, conflicts, branch_solver_prompt):
        """Verify that error message is given for non existing dimension"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_remove("idontexist")
        out, err = capsys.readouterr()
        assert "Dimension name 'idontexist' not found in conflicts" in out
        assert len(conflicts.get_resolved()) == 0

    def test_remove_dim_twice(self, capsys, conflicts, branch_solver_prompt):
        """Verify that error message is given trying to solve twice the same conflict"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_remove("missing")
        assert len(conflicts.get_resolved()) == 1
        branch_solver_prompt.do_remove("missing")
        out, err = capsys.readouterr()
        assert "Dimension name \'missing\' not found in conflicts" in out
        assert len(conflicts.get_resolved()) == 1

    def test_remove_dim_with_default(self, conflicts, branch_solver_prompt):
        """Verify that default value is properly parsed for remove resolution"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_remove("missing --default-value=9")
        assert len(conflicts.get_resolved()) == 1
        assert conflicts.get_resolved()[0].resolution.default_value == 9

    def test_remove_dim_with_bad_default(self, capsys, conflicts, branch_solver_prompt):
        """Verify error message is given when default value is invalid"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_remove("missing --default-value='bad'")
        out, err = capsys.readouterr()
        assert "could not convert string to float: 'bad'" in out
        assert len(conflicts.get_resolved()) == 0

    def test_remove_dim_with_cat(self, conflicts, branch_solver_prompt):
        """Verify that missing categorical dimension is properly removed"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_remove("missing_cat")
        assert len(conflicts.get_resolved()) == 1

    def test_remove_dim_with_cat_default(self, conflicts, branch_solver_prompt):
        """Verify that categorical default value is properly parsed for remove resolution"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_remove("missing_cat --default-value='goodbye'")
        assert len(conflicts.get_resolved()) == 1
        assert conflicts.get_resolved()[0].resolution.default_value == 'goodbye'

        conflicts.revert(conflicts.get_resolved()[0].resolution)
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_remove("missing_cat --default-value=5")
        assert len(conflicts.get_resolved()) == 1
        assert conflicts.get_resolved()[0].resolution.default_value == 5

    def test_remove_dim_with_cat_bad_default(self, capsys, conflicts, branch_solver_prompt):
        """Verify that error message is given for default value of invalid category"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_remove("missing_cat --default-value='bad'")
        out, err = capsys.readouterr()
        assert "Invalid category: bad" in out
        assert len(conflicts.get_resolved()) == 0

    def test_reset_remove(self, conflicts, branch_solver_prompt):
        """Verify that remove resolution is reverted"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_remove("missing")
        assert len(conflicts.get_resolved()) == 1
        branch_solver_prompt.do_reset("'{}'".format(str(conflicts.get_resolved()[0].resolution)))
        assert len(conflicts.get_resolved()) == 0

    def test_rename_dim(self, conflicts, branch_solver_prompt):
        """Verify that rename resolution is properly created"""
        assert len(conflicts.get_resolved()) == 0
        assert len(conflicts.get()) == 11
        branch_solver_prompt.do_rename("missing_idem new")
        assert len(conflicts.get_resolved()) == 2
        assert len(conflicts.get()) == 11

    def test_rename_and_change_dim(self, conflicts, branch_solver_prompt):
        """Verify that rename resolution is created and a new side-effect conflict added"""
        assert len(conflicts.get_resolved()) == 0
        assert len(conflicts.get()) == 11
        branch_solver_prompt.do_rename("missing new")
        assert len(conflicts.get_resolved()) == 2
        assert len(conflicts.get()) == 12

    def test_rename_bad_dim(self, capsys, conflicts, branch_solver_prompt):
        """Verify error messages when attempting invalid renamings"""
        assert len(conflicts.get_resolved()) == 0
        assert len(conflicts.get()) == 11
        branch_solver_prompt.do_rename("new_cat new")
        out, err = capsys.readouterr()
        assert "Dimension name \'new_cat\' not found in conflicts" in out
        assert len(conflicts.get_resolved()) == 0
        assert len(conflicts.get()) == 11

        assert len(conflicts.get_resolved()) == 0
        assert len(conflicts.get()) == 11
        branch_solver_prompt.do_rename("missing missing_cat")
        out, err = capsys.readouterr()
        assert "Dimension name \'missing_cat\' not found in conflicts" in out
        assert len(conflicts.get_resolved()) == 0
        assert len(conflicts.get()) == 11

    def test_reset_rename_with_same_priors(self, conflicts, branch_solver_prompt):
        """Verify that rename resolution is reverted"""
        assert len(conflicts.get_resolved()) == 0
        assert len(conflicts.get()) == 11
        branch_solver_prompt.do_rename("missing_idem new")
        assert len(conflicts.get_resolved()) == 2
        assert len(conflicts.get()) == 11
        branch_solver_prompt.do_reset("'{}'".format(str(conflicts.get_resolved()[0].resolution)))
        assert len(conflicts.get_resolved()) == 0
        assert len(conflicts.get()) == 11

    def test_reset_rename_with_different_priors(self, conflicts, branch_solver_prompt):
        """Verify that rename resolution is reverted and side-effect conflit is removed"""
        assert len(conflicts.get_resolved()) == 0
        assert len(conflicts.get()) == 11
        branch_solver_prompt.do_rename("missing new")
        assert len(conflicts.get_resolved()) == 2
        assert len(conflicts.get()) == 12
        branch_solver_prompt.do_reset("'{}'".format(str(conflicts.get_resolved()[0].resolution)))
        assert len(conflicts.get_resolved()) == 0
        assert len(conflicts.get()) == 11

    def test_set_code_change_type(self, conflicts, branch_solver_prompt):
        """Verify that code change resolution is created"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_code("break")
        assert len(conflicts.get_resolved()) == 1

    def test_set_code_change_bad_type(self, capsys, conflicts, branch_solver_prompt):
        """Verify error message when attempting code resolution with bad type"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_code("bad")
        out, err = capsys.readouterr()
        assert "invalid choice: 'bad'" in err
        assert len(conflicts.get_resolved()) == 0

    def test_set_code_change_type_twice(self, capsys, conflicts, branch_solver_prompt):
        """Verify that error message is given trying to solve twice the same conflict"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_code("break")
        assert len(conflicts.get_resolved()) == 1
        branch_solver_prompt.do_code("noeffect")
        out, err = capsys.readouterr()
        assert "No code conflicts to solve" in out
        assert len(conflicts.get_resolved()) == 1

    def test_reset_code(self, conflicts, branch_solver_prompt):
        """Verify that code resolution is reverted"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_code("break")
        assert len(conflicts.get_resolved()) == 1
        branch_solver_prompt.do_reset("'{}'".format(str(conflicts.get_resolved()[0].resolution)))
        assert len(conflicts.get_resolved()) == 0

    def test_set_algo(self, conflicts, branch_solver_prompt):
        """Verify that algo resolution is created"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_algo("")
        assert len(conflicts.get_resolved()) == 1

    def test_set_algo_twice(self, capsys, conflicts, branch_solver_prompt):
        """Verify that error message is given trying to solve twice the same conflict"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_algo("")
        assert len(conflicts.get_resolved()) == 1
        branch_solver_prompt.do_algo("")
        out, err = capsys.readouterr()
        assert "No algo conflict to solve" in out
        assert len(conflicts.get_resolved()) == 1

    def test_reset_algo(self, conflicts, branch_solver_prompt):
        """Verify that algo resolution is reverted"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_algo("")
        assert len(conflicts.get_resolved()) == 1
        branch_solver_prompt.do_reset("' {}'".format(str(conflicts.get_resolved()[0].resolution)))
        assert len(conflicts.get_resolved()) == 0

    def test_set_config_change_type(self, conflicts, branch_solver_prompt):
        """Verify that script's config resolution is created"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_config("break")
        assert len(conflicts.get_resolved()) == 1

    def test_set_config_change_bad_type(self, capsys, conflicts, branch_solver_prompt):
        """Verify error message when attempting config resolution with bad type"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_config("bad")
        out, err = capsys.readouterr()
        assert "invalid choice: 'bad'" in err
        assert len(conflicts.get_resolved()) == 0

    def test_set_config_change_type_twice(self, capsys, conflicts, branch_solver_prompt):
        """Verify that error message is given trying to solve twice the same conflict"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_config("break")
        assert len(conflicts.get_resolved()) == 1
        branch_solver_prompt.do_config("noeffect")
        out, err = capsys.readouterr()
        assert "No script's config conflicts to solve" in out
        assert len(conflicts.get_resolved()) == 1

    def test_reset_config(self, conflicts, branch_solver_prompt):
        """Verify that script's config resolution is reverted"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_config("break")
        assert len(conflicts.get_resolved()) == 1
        branch_solver_prompt.do_reset("'{}'".format(str(conflicts.get_resolved()[0].resolution)))
        assert len(conflicts.get_resolved()) == 0

    def test_set_commandline_change_type(self, conflicts, branch_solver_prompt):
        """Verify that cli resolution is created"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_commandline("break")
        assert len(conflicts.get_resolved()) == 1

    def test_set_commandline_change_bad_type(self, capsys, conflicts, branch_solver_prompt):
        """Verify error message when attempting cli resolution with bad type"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_commandline("bad")
        out, err = capsys.readouterr()
        assert "invalid choice: 'bad'" in err
        assert len(conflicts.get_resolved()) == 0

    def test_set_commandline_change_type_twice(self, capsys, conflicts, branch_solver_prompt):
        """Verify that error message is given trying to solve twice the same conflict"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_commandline("break")
        assert len(conflicts.get_resolved()) == 1
        branch_solver_prompt.do_commandline("noeffect")
        out, err = capsys.readouterr()
        assert "No command line conflicts to solve" in out
        assert len(conflicts.get_resolved()) == 1

    def test_reset_commandline(self, conflicts, branch_solver_prompt):
        """Verify that cli resolution is reverted"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_commandline("break")
        assert len(conflicts.get_resolved()) == 1
        branch_solver_prompt.do_reset("'{}'".format(str(conflicts.get_resolved()[0].resolution)))
        assert len(conflicts.get_resolved()) == 0

    def test_set_experiment_name(self, conflicts, branch_solver_prompt):
        """Verify that experiment name resolution is created"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_name("new-name")
        assert len(conflicts.get_resolved()) == 1

    def test_set_experiment_bad_name(self, capsys, conflicts, branch_solver_prompt):
        """Verify error message when attempting experiment name resolution with bad name"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_name("test")
        out, err = capsys.readouterr()
        assert "Experiment name 'test' already exist for version '1'" in out
        assert len(conflicts.get_resolved()) == 0

    def test_set_experiment_name_twice(self, capsys, conflicts, branch_solver_prompt):
        """Verify that error message is given trying to solve twice the same conflict"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_name("new-name")
        assert len(conflicts.get_resolved()) == 1
        branch_solver_prompt.do_name("whatever")
        out, err = capsys.readouterr()
        assert "No experiment name conflict to solve" in out
        assert len(conflicts.get_resolved()) == 1

    def test_reset_exp_name(self, conflicts, branch_solver_prompt):
        """Verify that experiment name resolution is reverted"""
        assert len(conflicts.get_resolved()) == 0
        branch_solver_prompt.do_name("new-name")
        assert len(conflicts.get_resolved()) == 1
        branch_solver_prompt.do_reset("'{}'".format(str(conflicts.get_resolved()[0].resolution)))
        assert len(conflicts.get_resolved()) == 0

    def test_commit_wont_quit_if_not_solved(self, conflicts, branch_solver_prompt):
        """Verify that commit will not quit if some conflicts are not resolved"""
        assert len(conflicts.get_resolved()) == 0
        assert len(conflicts.get()) == 11
        branch_solver_prompt.do_auto("")
        assert len(conflicts.get()) == 11
        assert len(conflicts.get_resolved()) == 10

        assert branch_solver_prompt.do_commit("") is False

    def test_commit_quit_if_solved(self, conflicts, branch_solver_prompt):
        """Verify that commit will quit when all conflicts are resolved"""
        branch_solver_prompt.do_auto("")
        conflicts.conflicts = [conflicts.get_resolved()[0]]
        assert conflicts.are_resolved
        assert branch_solver_prompt.do_commit("") is True

    def test_auto(self, conflicts, branch_solver_prompt):
        """Verify that all conflicts which requires not input are automatically resolved"""
        assert len(conflicts.get_resolved()) == 0
        assert len(conflicts.get()) == 11
        branch_solver_prompt.do_auto("")
        assert len(conflicts.get()) == 11
        assert len(conflicts.get_resolved()) == 10

    def test_reset_many(self, conflicts, branch_solver_prompt):
        """Verify that all resolutions are reverted"""
        assert len(conflicts.get_resolved()) == 0
        assert len(conflicts.get()) == 11
        branch_solver_prompt.do_auto("")
        assert len(conflicts.get()) == 11
        assert len(conflicts.get_resolved()) == 10

        reset_strings = []
        for resolution in conflicts.get_resolutions():
            resolution_string = shlex.quote((str(resolution)))
            # Otherwise --argument is interpreted as an argument by argparse rather than a
            # positional string
            if not resolution_string.startswith("'"):
                resolution_string = "' {}'".format(resolution_string)
            reset_strings.append(resolution_string)
        branch_solver_prompt.do_reset(" ".join(reset_strings))
        assert len(conflicts.get_resolved()) == 0



"""
:mod:`orion.plotting.base` -- Provides public plotting API
===========================================================

.. module:: base
   :platform: Unix
   :synopsis: Provides public plotting API

"""
import orion.plotting.backend_plotly as backend


def regret(experiment, order_by='suggested', verbose_hover=True, **kwargs):
    """
    Make a plot to visualize the performance of the hyper-optimization process.

    The x-axis contain the trials and the y-axis their respective best performance.

    Parameters
    ----------
    experiment: ExperimentClient, Experiment or ExperimentView
        The orion object containing the experiment data

    order_by: str
        Indicates how the trials should be ordered. Acceptable options are below.
        See attributes of ``Trial`` for more details.

        * 'suggested': Sort by trial suggested time (default).
        * 'reserved': Sort by trial reserved time.
        * 'completed': Sort by trial completed time.

    verbose_hover: bool
        Indicates whether to display the hyperparameter in hover tooltips. True by default.

    kwargs: dict
        All other plotting keyword arguments to be passed to
        :meth:`plotly.express.line`.

    Returns
    -------
    plotly.graph_objects.Figure

    Raises
    ------
    ValueError
        If no experiment is provided.

    """
    return backend.regret(experiment, order_by, verbose_hover, **kwargs)


PLOT_METHODS = {'regret': regret}


class PlotAccessor:
    """
    Make plots of ExperimentClient.

    Parameters
    ----------
    experiment : ExperimentClient
        The object for which the method is called.

    Raises
    ------
    ValueError
        If no experiment is provided.
    """

    def __init__(self, experiment):
        if not experiment:
            raise ValueError("Parameter 'experiment' is None")
        self._experiment = experiment

    def __call__(self, **kwargs):
        """
        Make different kinds of plots of ExperimentClient.

        Parameters
        ----------
        kind : str
            The kind of plot to produce:

            - 'regret' : Regret plot (default)
        """
        kind = kwargs.pop('kind', 'regret')

        if kind not in PLOT_METHODS.keys():
            raise ValueError(f"Plot of kind '{kind}' is not one of {list(PLOT_METHODS.keys())}")

        return PLOT_METHODS[kind](self._experiment, **kwargs)

    def regret(self, **kwargs):
        """Make a plot to visualize the performance of the hyper-optimization process."""
        __doc__ = regret.__doc__
        return self(kind="regret", **kwargs)



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.io.experiment_builder` -- Create experiment from user options
==============================================================================

.. module:: experiment
   :platform: Unix
   :synopsis: Functions which build `Experiment` and `ExperimentView` objects
       based on user configuration.


The instantiation of an `Experiment` is not a trivial process when the user request an experiment
with specific options. One can easily create a new experiment with
`ExperimentView('some_experiment_name')`, but the configuration of a _writable_ experiment is less
straighforward. This is because there is many sources of configuration and they have a strict
hierarchy. From the more global to the more specific, there is:

1. Global configuration:

  Defined by `orion.core.DEF_CONFIG_FILES_PATHS`.
  Can be scattered in user file system, defaults could look like:

    - `/some/path/to/.virtualenvs/orion/share/orion.core`
    - `/etc/xdg/xdg-ubuntu/orion.core`
    - `/home/${USER}/.config/orion.core`

  Note that some variables have default value even if user do not defined them in global
  configuration:

    - `max_trials = orion.core.io.resolve_config.DEF_CMD_MAX_TRIALS`
    - `pool_size = orion.core.io.resolve_config.DEF_CMD_POOL_SIZE`
    - `algorithms = random`
    - Database specific:

      * `database.name = 'orion'`
      * `database.type = 'MongoDB'`
      * `database.host = ${HOST}`

2. Oríon specific environment variables:

   Environment variables which can override global configuration

    - Database specific:

      * `ORION_DB_NAME`
      * `ORION_DB_TYPE`
      * `ORION_DB_ADDRESS`

3. Experiment configuration inside the database

  Configuration of the experiment if present in the database.
  Making this part of the configuration of the experiment makes it possible
  for the user to execute an experiment by only specifying partial configuration. The rest of the
  configuration is fetched from the database.

  For example, a user could:

    1. Rerun the same experiment

      Only providing the name is sufficient to rebuild the entire configuration of the
      experiment.

    2. Make a modification to an existing experiment

      The user can provide the name of the existing experiment and only provide the changes to
      apply on it. Here is an minimal example where we fully initialize a first experiment with a
      config file and then branch from it with minimal information.

      .. code-block:: bash

          # Initialize root experiment
          orion hunt --init-only --config previous_exeriment.yaml ./userscript -x~'uniform(0, 10)'
          # Branch a new experiment
          orion hunt -n previous_experiment ./userscript -x~'uniform(0, 100)'

4. Configuration file

  This configuration file is meant to overwrite the configuration coming from the database.
  If this configuration file was interpreted as part of the global configuration, a user could
  only modify an experiment using command line arguments.

5. Command-line arguments

  Those are the arguments provided to `orion` for any method (hunt, insert, etc). It includes the
  argument to `orion` itself as well as the user's script name and its arguments.

"""
import copy
import datetime
import getpass
import logging
import sys

from orion.algo.space import Space
import orion.core
from orion.core.evc.adapters import Adapter
from orion.core.evc.conflicts import detect_conflicts, ExperimentNameConflict
from orion.core.io import resolve_config
from orion.core.io.database import DuplicateKeyError
from orion.core.io.experiment_branch_builder import ExperimentBranchBuilder
from orion.core.io.interactive_commands.branching_prompt import BranchingPrompt
from orion.core.io.space_builder import SpaceBuilder
import orion.core.utils.backward as backward
from orion.core.utils.exceptions import (
    BranchingEvent, NoConfigurationError, NoNameError, RaceCondition)
from orion.core.worker.experiment import Experiment, ExperimentView
from orion.core.worker.primary_algo import PrimaryAlgo
from orion.core.worker.strategy import Strategy
from orion.storage.base import get_storage, setup_storage


log = logging.getLogger(__name__)


##
# Functions to build experiments
##

def build(name, version=None, branching=None, **config):
    """Build an experiment object

    If new, `space` argument must be provided, else all arguments are fetched from the database
    based on (name, version). If any argument given does not match the corresponding ones in the
    database for given (name, version), than the version is incremented and the experiment will be a
    child of the previous version.

    Parameters
    ----------
    name: str
        Name of the experiment to build
    version: int, optional
        Version to select. If None, last version will be selected. If version given is larger than
        largest version available, the largest version will be selected.
    space: dict, optional
        Optimization space of the algorithm. Should have the form `dict(name='<prior>(args)')`.
    algorithms: str or dict, optional
        Algorithm used for optimization.
    strategy: str or dict, optional
        Parallel strategy to use to parallelize the algorithm.
    max_trials: int, optional
        Maximum number or trials before the experiment is considered done.
    storage: dict, optional
        Configuration of the storage backend.
    branching: dict, optional
        Arguments to control the branching.

        branch_from: str, optional
            Name of the experiment to branch from.
        manual_resolution: bool, optional
            Starts the prompt to resolve manually the conflicts. Defaults to False.
        non_monitored_arguments: list of str, optional
            Will ignore these arguments while looking for differences. Defaults to [].
        ignore_code_changes: bool, optional
            Will ignore code changes while looking for differences. Defaults to False.
        algorithm_change: bool, optional
            Whether to automatically solve the algorithm conflict (change of algo config).
            Defaults to True.
        code_change_type: str, optional
            How to resolve code change automatically. Must be one of 'noeffect', 'unsure' or
            'break'.  Defaults to 'break'.
        cli_change_type: str, optional
            How to resolve cli change automatically. Must be one of 'noeffect', 'unsure' or 'break'.
            Defaults to 'break'.
        config_change_type: str, optional
            How to resolve config change automatically. Must be one of 'noeffect', 'unsure' or
            'break'.  Defaults to 'break'.

    """
    config = copy.deepcopy(config)
    for key, value in list(config.items()):
        if key.startswith('_') or value is None:
            config.pop(key)

    if 'strategy' in config:
        config['producer'] = {'strategy': config.pop('strategy')}

    if branching is None:
        branching = {}

    if branching.get('branch_from'):
        branching.setdefault('branch_to', name)
        name = branching['branch_from']

    db_config = fetch_config_from_db(name, version)

    new_config = config
    config = resolve_config.merge_configs(db_config, config)

    metadata = resolve_config.fetch_metadata(config.get('user'), config.get('user_args'))

    config = resolve_config.merge_configs(db_config, config, {'metadata': metadata})

    # TODO: Find a better solution
    if isinstance(config.get('algorithms'), dict) and len(config['algorithms']) > 1:
        config['algorithms'] = new_config['algorithms']

    config.setdefault('name', name)
    config.setdefault('version', version)

    if 'space' not in config:
        raise NoConfigurationError(
            'Experiment {} does not exist in DB and space was not defined.'.format(name))

    if len(config['space']) == 0:
        raise NoConfigurationError("No prior found. Please include at least one.")

    experiment = create_experiment(**copy.deepcopy(config))
    if experiment.id is None:
        try:
            _register_experiment(experiment)
        except DuplicateKeyError:
            experiment = build(branching=branching, **config)

        return experiment

    conflicts = _get_conflicts(experiment, branching)
    must_branch = len(conflicts.get()) > 1 or branching.get('branch_to')
    if must_branch:
        branched_experiment = _branch_experiment(experiment, conflicts, version, branching)
        try:
            _register_experiment(branched_experiment)
        except DuplicateKeyError as e:
            raise RaceCondition('There was a race condition during branching.') from e

        return branched_experiment

    _update_experiment(experiment)
    return experiment


def build_view(name, version=None):
    """Build experiment view

    An experiment view provides all reading operations of standard experiment but prevents the
    modification of the experiment and its trials.

    Parameters
    ----------
    name: str
        Name of the experiment to build
    version: int, optional
        Version to select. If None, last version will be selected. If version given is larger than
        largest version available, the largest version will be selected.

    """
    db_config = fetch_config_from_db(name, version)

    if not db_config:
        message = ("No experiment with given name '%s' and version '%s' inside database, "
                   "no view can be created." % (name, version if version else '*'))
        raise NoConfigurationError(message)

    db_config.setdefault('version', 1)

    experiment = create_experiment(**db_config)

    return ExperimentView(experiment)


def create_experiment(name, version, space, **kwargs):
    """Instantiate the experiment and its attribute objects

    All unspecified arguments will be replaced by system's defaults (orion.core.config.*).

    Parameters
    ----------
    name: str
        Name of the experiment.
    version: int
        Version of the experiment.
    space: dict or Space object
        Optimization space of the algorithm. If dict, should have the form
        `dict(name='<prior>(args)')`.
    algorithms: str or dict, optional
        Algorithm used for optimization.
    strategy: str or dict, optional
        Parallel strategy to use to parallelize the algorithm.
    max_trials: int, optional
        Maximum number or trials before the experiment is considered done.
    storage: dict, optional
        Configuration of the storage backend.

    """
    experiment = Experiment(name=name, version=version)
    experiment._id = kwargs.get('_id', None)  # pylint:disable=protected-access
    experiment.pool_size = kwargs.get('pool_size')
    if experiment.pool_size is None:
        experiment.pool_size = orion.core.config.experiment.get(
            'pool_size', deprecated='ignore')
    experiment.max_trials = kwargs.get('max_trials', orion.core.config.experiment.max_trials)
    experiment.space = _instantiate_space(space)
    experiment.algorithms = _instantiate_algo(experiment.space, kwargs.get('algorithms'))
    experiment.producer = kwargs.get('producer', {})
    experiment.producer['strategy'] = _instantiate_strategy(experiment.producer.get('strategy'))
    experiment.working_dir = kwargs.get('working_dir', orion.core.config.experiment.working_dir)
    experiment.metadata = kwargs.get('metadata', {'user': kwargs.get('user', getpass.getuser())})
    experiment.refers = kwargs.get('refers', {'parent_id': None, 'root_id': None, 'adapter': []})
    experiment.refers['adapter'] = _instantiate_adapters(experiment.refers.get('adapter', []))

    return experiment


def fetch_config_from_db(name, version=None):
    """Fetch configuration from database

    Parameters
    ----------
    name: str
        Name of the experiment to fetch
    version: int, optional
        Version to select. If None, last version will be selected. If version given is larger than
        largest version available, the largest version will be selected.

    """
    configs = get_storage().fetch_experiments({'name': name})

    if not configs:
        return {}

    config = _fetch_config_version(configs, version)

    if len(configs) > 1:
        log.info("Many versions for experiment %s have been found. Using latest "
                 "version %s.", name, config['version'])

    backward.populate_space(config, force_update=False)

    return config


##
# Private helper functions to build experiments
##

def _instantiate_adapters(config):
    """Instantiate the adapter object

    Parameters
    ----------
    config: list
         List of adapter configurations to build a CompositeAdapter for the EVC.

    """
    return Adapter.build(config)


def _instantiate_space(config):
    """Instantiate the space object

    Build the Space object if argument is a dictionary, else return the Space object as is.

    Parameters
    ----------
    config: dict or Space object
        Dictionary of priors or already built Space object.

    """
    if isinstance(config, Space):
        return config

    return SpaceBuilder().build(config)


def _instantiate_algo(space, config):
    """Instantiate the algorithm object

    Parameters
    ----------
    config: dict, optional
        Configuration of the strategy. If None of empty, system's defaults are used
        (orion.core.config.experiment.algorithms).

    """
    if not config:
        config = orion.core.config.experiment.algorithms

    return PrimaryAlgo(space, config)


def _instantiate_strategy(config=None):
    """Instantiate the strategy object

    Parameters
    ----------
    config: dict, optional
        Configuration of the strategy. If None of empty, system's defaults are used
        (orion.core.config.producer.strategy).

    """
    if not config:
        config = orion.core.config.experiment.strategy

    if isinstance(config, str):
        strategy_type = config
        config = {}
    else:
        strategy_type, config = next(iter(config.items()))

    return Strategy(of_type=strategy_type, **config)


def _register_experiment(experiment):
    """Register a new experiment in the database"""
    experiment.metadata['datetime'] = datetime.datetime.utcnow()
    config = experiment.configuration
    # This will raise DuplicateKeyError if a concurrent experiment with
    # identical (name, metadata.user) is written first in the database.

    get_storage().create_experiment(config)

    # XXX: Reminder for future DB implementations:
    # MongoDB, updates an inserted dict with _id, so should you :P
    experiment._id = config['_id']  # pylint:disable=protected-access

    # Update refers in db if experiment is root
    if experiment.refers.get('parent_id') is None:
        log.debug('update refers (name: %s)', experiment.name)
        experiment.refers['root_id'] = experiment.id
        get_storage().update_experiment(experiment, refers=experiment.configuration['refers'])


def _update_experiment(experiment):
    """Update experiment configuration in database"""
    log.debug('updating experiment (name: %s)', experiment.name)
    config = experiment.configuration

    # TODO: Remove since this should not occur anymore without metadata.user in the indices?
    # Writing the final config to an already existing experiment raises
    # a DuplicatKeyError because of the embedding id `metadata.user`.
    # To avoid this `final_config["name"]` is popped out before
    # `db.write()`, thus seamingly breaking  the compound index
    # `(name, metadata.user)`
    config.pop("name")

    get_storage().update_experiment(experiment, **config)


def _branch_experiment(experiment, conflicts, version, branching_arguments):
    """Create a new branch experiment with adapters for the given conflicts"""
    experiment_brancher = ExperimentBranchBuilder(conflicts, **branching_arguments)

    needs_manual_resolution = (not experiment_brancher.is_resolved or
                               experiment_brancher.manual_resolution)

    if not experiment_brancher.is_resolved:
        name_conflict = conflicts.get([ExperimentNameConflict])[0]
        if not name_conflict.is_resolved and not version:
            raise RaceCondition('There was likely a race condition during version increment.')

    if needs_manual_resolution:
        # TODO: This should only be possible when using cmdline API
        branching_prompt = BranchingPrompt(experiment_brancher)

        if not sys.__stdin__.isatty():
            raise BranchingEvent(branching_prompt.get_status())

        branching_prompt.cmdloop()

        if branching_prompt.abort or not experiment_brancher.is_resolved:
            sys.exit()

    config = experiment_brancher.conflicting_config
    config['refers']['adapter'] = experiment_brancher.create_adapters().configuration
    config['refers']['parent_id'] = experiment.id

    config.pop('_id')

    return create_experiment(**config)


def _get_conflicts(experiment, branching):
    """Get conflicts between current experiment and corresponding configuration in database"""
    db_experiment = build_view(experiment.name, experiment.version)
    conflicts = detect_conflicts(db_experiment.configuration, experiment.configuration,
                                 branching)

    # elif must_branch and not enable_branching:
    #     raise ValueError("Configuration is different and generate a branching event")

    return conflicts


def _fetch_config_version(configs, version=None):
    """Fetch the experiment configuration corresponding to the given version

    Parameters
    ----------
    configs: list
        List of configurations fetched from storoge.
    version: int, optional
        Version to select. If None, last version will be selected. If version given is larger than
        largest version available, the largest version will be selected.

    """
    max_version = max(configs, key=lambda exp: exp.get('version', 1)).get('version', 1)

    if version is None:
        version = max_version
    else:
        version = version

    if version > max_version:
        log.warning("Version %s was specified but most recent version is only %s. "
                    "Using %s.", version, max_version, max_version)

    version = min(version, max_version)

    configs = filter(lambda exp: exp.get('version', 1) == version, configs)

    return next(iter(configs))


###
# Functions for commandline API
###

def build_from_args(cmdargs):
    """Build an experiment based on commandline arguments.

    Options provided in commandline and configuration file (--config) will overwrite system's
    default values and configuration from database if experiment already exits.
    Commandline arguments have precedence over configuration file options.

    .. seealso::

        :func:`orion.core.io.experiment_builder.build` for more information on experiment creation.

    """
    cmd_config = get_cmd_config(cmdargs)

    if 'name' not in cmd_config:
        raise NoNameError()

    setup_storage(cmd_config['storage'], debug=cmd_config.get('debug'))

    return build(**cmd_config)


def build_view_from_args(cmdargs):
    """Build an experiment view based on commandline arguments

    .. seealso::

        :func:`orion.core.io.experiment_builder.build_view` for more information on experiment view
        creation.

    """
    cmd_config = get_cmd_config(cmdargs)

    if 'name' not in cmd_config:
        raise NoNameError()

    setup_storage(cmd_config['storage'], debug=cmd_config.get('debug'))

    name = cmd_config.get('name')
    version = cmd_config.get('version')

    return build_view(name, version)


def get_cmd_config(cmdargs):
    """Fetch configuration defined by commandline and local configuration file.

    Arguments of commandline have priority over options in configuration file.
    """
    cmdargs = resolve_config.fetch_config_from_cmdargs(cmdargs)
    cmd_config = resolve_config.fetch_config(cmdargs)
    cmd_config = resolve_config.merge_configs(cmd_config, cmdargs)

    cmd_config.update(cmd_config.pop('experiment', {}))
    cmd_config['branching'] = cmd_config.pop('evc', {})

    # TODO: user_args won't be defined if reading from DB only (`orion hunt -n <exp> ` alone)
    metadata = resolve_config.fetch_metadata(cmd_config.get('user'), cmd_config.get('user_args'))
    cmd_config['metadata'] = metadata
    cmd_config.pop('config', None)

    backward.populate_space(cmd_config)
    backward.update_db_config(cmd_config)

    return cmd_config



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.worker.trial`."""
import bson
import numpy
import pytest

from orion.core.worker.trial import Trial


class TestTrial(object):
    """Test Trial object and class."""

    def test_init_empty(self):
        """Initialize empty trial."""
        t = Trial()
        assert t.experiment is None
        assert t.status == 'new'
        assert t.worker is None
        assert t.submit_time is None
        assert t.start_time is None
        assert t.end_time is None
        assert t.results == []
        assert t.params == {}
        assert t.working_dir is None

    def test_init_full(self, exp_config):
        """Initialize with a dictionary with complete specification."""
        t = Trial(**exp_config[1][1])
        assert t.experiment == exp_config[1][1]['experiment']
        assert t.status == exp_config[1][1]['status']
        assert t.worker == exp_config[1][1]['worker']
        assert t.submit_time == exp_config[1][1]['submit_time']
        assert t.start_time == exp_config[1][1]['start_time']
        assert t.end_time == exp_config[1][1]['end_time']
        assert list(map(lambda x: x.to_dict(), t.results)) == exp_config[1][1]['results']
        assert t.results[0].name == exp_config[1][1]['results'][0]['name']
        assert t.results[0].type == exp_config[1][1]['results'][0]['type']
        assert t.results[0].value == exp_config[1][1]['results'][0]['value']
        assert list(map(lambda x: x.to_dict(), t._params)) == exp_config[1][1]['params']
        assert t.working_dir is None

    def test_higher_shapes_not_ndarray(self):
        """Test that `numpy.ndarray` values are converted to list."""
        value = numpy.zeros([3])
        expected = value.tolist()
        params = [dict(name='/x', type='real', value=value)]
        trial = Trial(params=params)

        assert trial._params[0].value == expected

    def test_bad_access(self):
        """Other than `Trial.__slots__` are not allowed."""
        t = Trial()
        with pytest.raises(AttributeError):
            t.asdfa = 3

    def test_bad_init(self):
        """Other than `Trial.__slots__` are not allowed in __init__ too."""
        with pytest.raises(AttributeError):
            Trial(ispii='iela')

    def test_value_bad_init(self):
        """Other than `Trial.Value.__slots__` are not allowed in __init__ too."""
        with pytest.raises(AttributeError):
            Trial.Value(ispii='iela')

    def test_param_bad_init(self):
        """Other than `Trial.Param.__slots__` are not allowed in __init__ too."""
        with pytest.raises(AttributeError):
            Trial.Param(ispii='iela')

    def test_result_bad_init(self):
        """Other than `Trial.Result.__slots__` are not allowed in __init__ too."""
        with pytest.raises(AttributeError):
            Trial.Result(ispii='iela')

    def test_not_allowed_status(self):
        """Other than `Trial.allowed_stati` are not allowed in `Trial.status`."""
        t = Trial()
        with pytest.raises(ValueError):
            t.status = 'asdf'
        with pytest.raises(ValueError):
            t = Trial(status='ispi')

    def test_value_not_allowed_type(self):
        """Other than `Trial.Result.allowed_types` are not allowed in `Trial.Result.type`.

        Same for `Trial.Param`.
        """
        with pytest.raises(ValueError):
            v = Trial.Result(name='asfda', type='hoho')
        v = Trial.Result()
        with pytest.raises(ValueError):
            v.type = 'asfda'

        with pytest.raises(ValueError):
            v = Trial.Param(name='asfda', type='hoho')
        v = Trial.Param()
        with pytest.raises(ValueError):
            v.type = 'asfda'

    @pytest.mark.usefixtures("clean_db")
    def test_convertion_to_dict(self, exp_config):
        """Convert to dictionary form for database using ``dict``."""
        t = Trial(**exp_config[1][1])
        assert t.to_dict() == exp_config[1][1]

    @pytest.mark.usefixtures("clean_db")
    def test_build_trials(self, exp_config):
        """Convert to objects form using `Trial.build`."""
        trials = Trial.build(exp_config[1])
        assert list(map(lambda x: x.to_dict(), trials)) == exp_config[1]

    @pytest.mark.usefixtures("clean_db")
    def test_value_equal(self, exp_config):
        """Compare Param objects using __eq__"""
        trials = Trial.build(exp_config[1])

        assert trials[0]._params[0] == Trial.Param(**exp_config[1][0]['params'][0])
        assert trials[0]._params[1] != Trial.Param(**exp_config[1][0]['params'][0])

    def test_str_trial(self, exp_config):
        """Test representation of `Trial`."""
        t = Trial(**exp_config[1][1])
        assert str(t) == "Trial(experiment='supernaedo2-dendi', status='completed', " \
                         "params=/decoding_layer:lstm_with_attention,/encoding_layer:gru)"

    def test_str_value(self, exp_config):
        """Test representation of `Trial.Value`."""
        t = Trial(**exp_config[1][1])
        assert (str(t._params[1]) ==
                "Param(name='/encoding_layer', type='categorical', value='gru')")

    def test_invalid_result(self, exp_config):
        """Test that invalid objectives cannot be set"""
        t = Trial(**exp_config[1][1])

        # Make sure valid ones pass
        t.results = [Trial.Result(name='asfda', type='constraint', value=None),
                     Trial.Result(name='asfdb', type='objective', value=1e-5)]

        # No results at all
        with pytest.raises(ValueError) as exc:
            t.results = []
        assert 'No objective found' in str(exc.value)

        # No objectives
        with pytest.raises(ValueError) as exc:
            t.results = [Trial.Result(name='asfda', type='constraint', value=None)]
        assert 'No objective found' in str(exc.value)

        # Bad objective type
        with pytest.raises(ValueError) as exc:
            t.results = [Trial.Result(name='asfda', type='constraint', value=None),
                         Trial.Result(name='asfdb', type='objective', value=None)]
        assert 'Results must contain' in str(exc.value)

    def test_objective_property(self):
        """Check property `Trial.objective`."""
        base_trial = {
            'results': []
        }
        base_trial['results'].append({'name': 'a', 'type': 'gradient', 'value': 0.5})

        # 0 objective in `results` list
        trial = Trial(**base_trial)

        assert trial.objective is None

        # 1 results in `results` list
        expected = Trial.Result(name='b', type='objective', value=42)

        base_trial['results'].append({'name': 'b', 'type': 'objective', 'value': 42})
        trial = Trial(**base_trial)

        assert trial.objective == expected

        # >1 results in `results` list
        base_trial['results'].append({'name': 'c', 'type': 'objective', 'value': 42})
        trial = Trial(**base_trial)

        assert trial.objective == expected

    def test_gradient_property(self):
        """Check property `Trial.gradient`."""
        base_trial = {
            'results': []
        }

        # 0 result exist
        trial = Trial(**base_trial)
        assert trial.gradient is None

        # 0 result of type 'gradient' exist
        base_trial['results'].append({'name': 'a', 'type': 'objective', 'value': 10})
        trial = Trial(**base_trial)

        assert trial.gradient is None

        # 1 result of type 'gradient' exist
        expected = Trial.Result(name='b', type='gradient', value=5)

        base_trial['results'].append({'name': 'b', 'type': 'gradient', 'value': 5})
        trial = Trial(**base_trial)

        assert trial.gradient == expected

        # >1 gradient result
        base_trial['results'].append({'name': 'c', 'type': 'gradient', 'value': [12, 15]})
        trial = Trial(**base_trial)

        assert trial.gradient == expected

    def test_constraints_property(self):
        """Tests the property for accessing constraints"""
        base_trial = {
            'results': []
        }

        # 0 result exist
        trial = Trial(**base_trial)
        assert trial.constraints == []

        # 0 result of type 'constraints' exist
        base_trial['results'].append({'name': 'a', 'type': 'objective', 'value': 10})
        trial = Trial(**base_trial)

        assert trial.constraints == []

        # 1 result of type 'constraint' exist
        expected = [
            Trial.Result(name='b', type='constraint', value=5)
        ]

        base_trial['results'].append({'name': 'b', 'type': 'constraint', 'value': 5})
        trial = Trial(**base_trial)

        assert expected == trial.constraints

        # > 1 results of type 'constraint' exist
        expected = [
            Trial.Result(name='b', type='constraint', value=5),
            Trial.Result(name='c', type='constraint', value=20)
        ]

        base_trial['results'].append({'name': 'c', 'type': 'constraint', 'value': 20})
        trial = Trial(**base_trial)

        assert expected == trial.constraints

    def test_statistics_property(self):
        """Tests the property for accessing statistics"""
        base_trial = {
            'results': []
        }

        # 0 result exist
        trial = Trial(**base_trial)
        assert trial.statistics == []

        # 0 result of type 'statistic' exist
        base_trial['results'].append({'name': 'a', 'type': 'objective', 'value': 10})
        trial = Trial(**base_trial)

        assert trial.statistics == []

        # 1 result of type 'statistic' exist
        expected = [
            Trial.Result(name='b', type='statistic', value=5)
        ]

        base_trial['results'].append({'name': 'b', 'type': 'statistic', 'value': 5})
        trial = Trial(**base_trial)

        assert expected == trial.statistics

        # > 1 results of type 'statistic' exist
        expected = [
            Trial.Result(name='b', type='statistic', value=5),
            Trial.Result(name='c', type='statistic', value=20)
        ]

        base_trial['results'].append({'name': 'c', 'type': 'statistic', 'value': 20})
        trial = Trial(**base_trial)

        assert expected == trial.statistics

    def test_params_repr_property(self, exp_config):
        """Check property `Trial.params_repr`."""
        t = Trial(**exp_config[1][1])
        assert Trial.format_params(t._params) == \
               "/decoding_layer:lstm_with_attention,/encoding_layer:gru"
        assert Trial.format_params(t._params, sep='\n') == \
               "/decoding_layer:lstm_with_attention\n/encoding_layer:gru"

        t = Trial()
        assert Trial.format_params(t._params) == ""

    def test_hash_name_property(self, exp_config):
        """Check property `Trial.hash_name`."""
        t = Trial(**exp_config[1][1])
        assert t.hash_name == "ebcf6c6c8604f96444af1c3e519aea7f"

        t = Trial()
        with pytest.raises(ValueError) as exc:
            t.hash_name
        assert 'params' in str(exc.value)

    def test_param_name_property(self, exp_config):
        """Check property `Trial.hash_params`."""
        exp_config[1][1]['params'].append({'name': '/max_epoch', 'type': 'fidelity', 'value': '1'})
        t1 = Trial(**exp_config[1][1])
        exp_config[1][1]['params'][-1]['value'] = '2'  # changing the fidelity
        t2 = Trial(**exp_config[1][1])
        assert t1.hash_name != t2.hash_name
        assert t1.hash_params == t2.hash_params

    def test_hash_ignore_experiment(self, exp_config):
        """Check property `Trial.compute_trial_hash(ignore_experiment=True)`."""
        exp_config[1][1]['params'].append({'name': '/max_epoch', 'type': 'fidelity', 'value': '1'})
        t1 = Trial(**exp_config[1][1])
        exp_config[1][1]['experiment'] = 'test'  # changing the experiment name
        t2 = Trial(**exp_config[1][1])
        assert t1.hash_name != t2.hash_name
        assert t1.hash_params != t2.hash_params
        assert (Trial.compute_trial_hash(t1, ignore_experiment=True) ==
                Trial.compute_trial_hash(t2, ignore_experiment=True))

    def test_hash_ignore_lie(self, exp_config):
        """Check property `Trial.compute_trial_hash(ignore_lie=True)`."""
        exp_config[1][1]['params'].append({'name': '/max_epoch', 'type': 'fidelity', 'value': '1'})
        t1 = Trial(**exp_config[1][1])
        # Add a lie
        exp_config[1][1]['results'].append({'name': 'lie', 'type': 'lie', 'value': 1})
        t2 = Trial(**exp_config[1][1])
        assert t1.hash_name != t2.hash_name
        assert t1.hash_params == t2.hash_params
        assert (Trial.compute_trial_hash(t1, ignore_lie=True) ==
                Trial.compute_trial_hash(t2, ignore_lie=True))

    def test_full_name_property(self, exp_config):
        """Check property `Trial.full_name`."""
        t = Trial(**exp_config[1][1])
        assert t.full_name == ".decoding_layer:lstm_with_attention-.encoding_layer:gru"

        t = Trial()
        with pytest.raises(ValueError) as exc:
            t.full_name
        assert 'params' in str(exc.value)

    def test_higher_shape_id_is_same(self):
        """Check if a Trial with a shape > 1 has the same id once it has been through the DB."""
        x = {'name': '/x', 'value': [1, 2], 'type': 'real'}
        trial = Trial(params=[x])
        assert trial.id == Trial(**bson.BSON.decode(bson.BSON.encode(trial.to_dict()))).id



"""Test for generic :class:`orion.core.evc.tree`"""

from orion.core.evc.tree import DepthFirstTraversal, flattened, PreOrderTraversal, TreeNode


def test_node_creation():
    """Test empty initialization of tree node"""
    TreeNode("test")
    TreeNode("test", None, tuple())


def test_node_creation_with_parent():
    """Test assignement of parent on initialization"""
    parent = TreeNode("test")
    child = TreeNode("test", parent)

    assert child.parent is parent

    assert parent.children[0] is child


def test_node_set_parent():
    """Test assignement of parent"""
    parent = TreeNode("test")
    child = TreeNode("test")

    child.set_parent(parent)

    assert child.parent is parent

    assert parent.children[0] is child


def test_node_set_no_parent():
    """Test assignement of no parent, meaning droping the parent"""
    parent = TreeNode("test")
    child = TreeNode("test")

    child.set_parent(parent)

    assert child.parent is parent
    assert parent.children[0] is child

    child.set_parent(None)

    assert child.parent is None
    assert len(parent.children) == 0


def test_node_drop_parent():
    """Test dropping parent"""
    parent = TreeNode("test")
    child = TreeNode("test")

    child.set_parent(parent)

    assert child.parent is parent
    assert parent.children[0] is child

    child.drop_parent()

    assert child.parent is None
    assert len(parent.children) == 0


def test_node_add_child():
    """Test assignement of child"""
    parent = TreeNode("test")
    child = TreeNode("test")

    parent.add_children(child)

    assert parent.children[0] is child
    assert child.parent is parent


def test_node_remove_child():
    """Test removing child"""
    parent = TreeNode("test")
    child = TreeNode("test")

    parent.add_children(child)

    assert child.parent is parent
    assert parent.children[0] is child

    parent.drop_children(child)

    assert len(parent.children) == 0
    assert child.parent is None


def test_node_add_children():
    """Test assignement of children"""
    parent = TreeNode("test")
    child1 = TreeNode("test1")
    child2 = TreeNode("test2")

    parent.add_children(child1, child2)

    assert parent.children[0] is child1
    assert parent.children[1] is child2
    assert child1.parent is parent
    assert child2.parent is parent


def test_node_remove_children_sequentially():
    """Test removing children one after the other"""
    parent = TreeNode("test")
    child1 = TreeNode("test1")
    child2 = TreeNode("test2")

    parent.add_children(child1, child2)

    assert parent.children[0] is child1
    assert parent.children[1] is child2
    assert child1.parent is parent
    assert child2.parent is parent

    parent.drop_children(child1)

    assert len(parent.children) == 1
    assert child1.parent is None
    assert parent.children[0] is child2
    assert child2.parent is parent

    parent.drop_children(child2)

    assert len(parent.children) == 0
    assert child1.parent is None
    assert child2.parent is None


def test_node_remove_children_in_batch():
    """Test removing children all at the same time"""
    parent = TreeNode("test")
    child1 = TreeNode("test1")
    child2 = TreeNode("test2")

    parent.add_children(child1, child2)

    assert parent.children[0] is child1
    assert parent.children[1] is child2
    assert child1.parent is parent
    assert child2.parent is parent

    parent.drop_children(child1, child2)

    assert len(parent.children) == 0
    assert child1.parent is None
    assert child2.parent is None


def test_node_remove_all_children():
    """Test drop_children() drops all of them"""
    parent = TreeNode("test")
    child1 = TreeNode("test1")
    child2 = TreeNode("test2")

    parent.add_children(child1, child2)

    assert parent.children[0] is child1
    assert parent.children[1] is child2
    assert child1.parent is parent
    assert child2.parent is parent

    parent.drop_children()

    assert len(parent.children) == 0
    assert child1.parent is None
    assert child2.parent is None


def test_parent_parent():
    """Test path through two level of parents"""
    grand_parent = TreeNode("test")
    parent = TreeNode("test", grand_parent)
    child = TreeNode("test", parent)

    assert child.parent is parent
    assert child.parent.parent is grand_parent
    assert grand_parent.children[0] is parent
    assert grand_parent.children[0].children[0] is child


def test_children_children():
    """Test path through two level of children"""
    child = TreeNode("test")
    parent = TreeNode("test", children=[child])
    grand_parent = TreeNode("test", children=[parent])

    assert child.parent is parent
    assert child.parent.parent is grand_parent
    assert grand_parent.children[0] is parent
    assert grand_parent.children[0].children[0] is child


def test_root():
    """Test root access from leaf"""
    child = TreeNode("test")
    parent = TreeNode("test", children=[child])
    grand_parent = TreeNode("test", children=[parent])

    assert child.root is grand_parent
    assert child.root is parent.root
    assert grand_parent.root is grand_parent


def test_node_parent_reinsertion():
    """Test that replacing a parent with the same one does not change anything"""
    parent = TreeNode("test")
    child1 = TreeNode("test", parent)
    child2 = TreeNode("test", parent)

    assert child1.parent is parent
    assert child2.parent is parent
    assert parent.children[0] is child1
    assert parent.children[1] is child2

    child1.set_parent(parent)

    assert child1.parent is parent
    assert child2.parent is parent
    assert parent.children[0] is child1
    assert parent.children[1] is child2

    child2.set_parent(parent)

    assert child1.parent is parent
    assert child2.parent is parent
    assert parent.children[0] is child1
    assert parent.children[1] is child2


def test_node_children_reinsertion():
    """Test that replacing a children with the same one does not change anything"""
    parent = TreeNode("test")
    child1 = TreeNode("test", parent)
    child2 = TreeNode("test", parent)

    assert child1.parent is parent
    assert child2.parent is parent
    assert parent.children[0] is child1
    assert parent.children[1] is child2

    parent.add_children(child1)

    assert child1.parent is parent
    assert child2.parent is parent
    assert parent.children[0] is child1
    assert parent.children[1] is child2

    parent.add_children(child2)

    assert child1.parent is parent
    assert child2.parent is parent
    assert parent.children[0] is child1
    assert parent.children[1] is child2


def test_node_set_parent_drop_children():
    """Test that replacing a parent removes the node from parent's children"""
    grand_parent = TreeNode("test")
    parent = TreeNode("test", grand_parent)
    child = TreeNode("test", parent)

    assert child.parent is parent
    assert child.parent.parent is grand_parent
    assert grand_parent.children[0] is parent
    assert grand_parent.children[0].children[0] is child

    # That's silly...
    child.set_parent(grand_parent)

    assert child.parent is grand_parent
    assert len(grand_parent.children) == 2
    assert grand_parent.parent is None
    assert len(parent.children) == 0
    assert parent.parent is grand_parent


def test_node_add_children_drop_children():
    """Test that replacing a children removes the node from previous parent's children"""
    grand_parent = TreeNode("test")
    parent = TreeNode("test", grand_parent)
    child = TreeNode("test", parent)

    assert child.parent is parent
    assert child.parent.parent is grand_parent
    assert grand_parent.children[0] is parent
    assert grand_parent.children[0].children[0] is child

    # That's silly...
    grand_parent.add_children(child)

    assert child.parent is grand_parent
    assert len(grand_parent.children) == 2
    assert grand_parent.parent is None
    assert len(parent.children) == 0
    assert parent.parent is grand_parent


def test_preorder_traversal():
    """Test order retrieval of DepthFirstTraversal"""
    # a
    # |   \  \   \
    # b    c  d   e
    # | \         |
    # f  g        h
    root = TreeNode("a")
    b = TreeNode("b", root)
    TreeNode("c", root)
    TreeNode("d", root)
    e = TreeNode("e", root)

    TreeNode("f", b)
    TreeNode("g", b)

    TreeNode("h", e)

    rval = [node.item for node in PreOrderTraversal(root)]
    assert rval == ['a', 'b', 'f', 'g', 'c', 'd', 'e', 'h']


def test_depth_first_traversal():
    """Test order retrieval of DepthFirstTraversal"""
    # a
    # |   \  \   \
    # b    c  d   e
    # | \         |
    # f  g        h
    root = TreeNode("a")
    b = TreeNode("b", root)
    TreeNode("c", root)
    TreeNode("d", root)
    e = TreeNode("e", root)

    TreeNode("f", b)
    TreeNode("g", b)

    TreeNode("h", e)

    rval = [node.item for node in DepthFirstTraversal(root)]
    assert rval == ['f', 'g', 'b', 'c', 'd', 'h', 'e', 'a']


def test_map_children():
    """Test recursive application of map on children"""
    # a
    # |   \  \   \
    # b    c  d   e
    # | \         |
    # f  g        h
    #
    # Should become
    #
    # 2
    # |   \  \   \
    # 3    3  3   3
    # | \         |
    # 4  4        4

    root = TreeNode(1)
    b = TreeNode(1, root)
    TreeNode(1, root)
    TreeNode(1, root)
    e = TreeNode(1, root)

    TreeNode(1, b)
    TreeNode(1, b)

    TreeNode(1, e)

    def increment(node, children):
        for child in TreeNode(0, None, children):
            child.item += 1

        return node.item + 1, children

    rval = root.map(increment, root.children)
    assert [node.item for node in rval] == [2, 3, 4, 4, 3, 3, 3, 4]


def test_map_parent():
    """Test recursive application of map on parents"""
    # a
    # |   \  \   \
    # b    c  d   e
    # | \         |
    # f  g        h
    #
    # Will give
    #
    # 4
    # |
    # 3
    # |
    # 2

    root = TreeNode(1)
    b = TreeNode(1, root)
    TreeNode(1, root)
    TreeNode(1, root)
    e = TreeNode(1, root)

    f = TreeNode(1, b)
    TreeNode(1, b)

    h = TreeNode(1, e)

    def increment_parent(node, parent):
        if parent is not None:
            for parent in parent.root:
                parent.item += 1

        return node.item + 1, parent

    rval = f.map(increment_parent, f.parent)
    assert [node.item for node in rval.root] == [4, 3, 2]

    rval = h.map(increment_parent, h.parent)
    assert [node.item for node in rval.root] == [4, 3, 2]


def test_flattened():
    """Test flattened tree into a list, retrieving items"""
    # a
    # |   \  \   \
    # b    c  d   e
    # | \         |
    # f  g        h
    root = TreeNode("a")
    b = TreeNode("b", root)
    TreeNode("c", root)
    TreeNode("d", root)
    e = TreeNode("e", root)

    TreeNode("f", b)
    TreeNode("g", b)

    TreeNode("h", e)

    assert flattened(root) == ['a', 'b', 'f', 'g', 'c', 'd', 'e', 'h']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.cli.test_db` -- Module to check if the DB works
=================================================================

.. module:: test_db
   :platform: Unix
   :synopsis: Runs multiple checks to see if the database was correctly setup.

"""
import argparse
import logging

from orion.core.cli.db.test import main


log = logging.getLogger(__name__)
DESCRIPTION = '(DEPRECATED) Use command `orion db test` instead'


def add_subparser(parser):
    """Add the subparser that needs to be used for this command"""
    test_db_parser = parser.add_parser('test-db', help=DESCRIPTION, description=DESCRIPTION)

    test_db_parser.add_argument('-c', '--config', type=argparse.FileType('r'),
                                metavar='path-to-config', help="user provided "
                                "orion configuration file")

    test_db_parser.set_defaults(func=wrap_main)

    return test_db_parser


def wrap_main(args):
    """Run through all checks for database."""
    log.warning('Command `orion test-db` is deprecated and will be removed in v0.2.0. Use '
                '`orion db test` instead.')

    main(args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example usage and tests for :mod:`orion.client`."""
import copy
from importlib import reload
import json

import pytest

import orion.client
from orion.client import get_experiment
import orion.client.cli as cli
import orion.core
from orion.core.io.database.ephemeraldb import EphemeralDB
from orion.core.io.database.pickleddb import PickledDB
from orion.core.utils.exceptions import BranchingEvent, NoConfigurationError, RaceCondition
from orion.core.utils.singleton import SingletonNotInstantiatedError, update_singletons
from orion.core.worker.experiment import ExperimentView
from orion.storage.base import get_storage
from orion.storage.legacy import Legacy
from orion.testing import OrionState

create_experiment = orion.client.create_experiment
workon = orion.client.workon


config = dict(
    name='supernaekei',
    space={'x': 'uniform(0, 200)'},
    metadata={'user': 'tsirif',
              'orion_version': 'XYZ',
              'VCS': {"type": "git",
                      "is_dirty": False,
                      "HEAD_sha": "test",
                      "active_branch": None,
                      "diff_sha": "diff"}},
    version=1,
    pool_size=1,
    max_trials=10,
    working_dir='',
    algorithms={'random': {'seed': 1}},
    producer={'strategy': 'NoParallelStrategy'},
    refers=dict(
        root_id='supernaekei',
        parent_id=None,
        adapter=[])
    )


@pytest.fixture()
def user_config():
    """Curate config as a user would provide it"""
    user_config = copy.deepcopy(config)
    user_config.pop('metadata')
    user_config.pop('version')
    user_config['strategy'] = user_config.pop('producer')['strategy']
    user_config.pop('refers')
    user_config.pop('pool_size')
    return user_config


@pytest.fixture()
def data():
    """Return serializable data."""
    return "this is datum"


class TestReportResults(object):
    """Check functionality and edge cases of `report_results` helper interface."""

    def test_with_no_env(self, monkeypatch, capsys, data):
        """Test without having set the appropriate environmental variable.

        Then: It should print `data` parameter instead to stdout.
        """
        monkeypatch.delenv('ORION_RESULTS_PATH', raising=False)
        reloaded_client = reload(cli)

        assert reloaded_client.IS_ORION_ON is False
        assert reloaded_client.RESULTS_FILENAME is None
        assert reloaded_client._HAS_REPORTED_RESULTS is False

        reloaded_client.report_results(data)
        out, err = capsys.readouterr()
        assert reloaded_client._HAS_REPORTED_RESULTS is True
        assert out == data + '\n'
        assert err == ''

    def test_with_correct_env(self, monkeypatch, capsys, tmpdir, data):
        """Check that a file with correct data will be written to an existing
        file in a legit path.
        """
        path = str(tmpdir.join('naedw.txt'))
        with open(path, mode='w'):
            pass
        monkeypatch.setenv('ORION_RESULTS_PATH', path)
        reloaded_client = reload(cli)

        assert reloaded_client.IS_ORION_ON is True
        assert reloaded_client.RESULTS_FILENAME == path
        assert reloaded_client._HAS_REPORTED_RESULTS is False

        reloaded_client.report_results(data)
        out, err = capsys.readouterr()
        assert reloaded_client._HAS_REPORTED_RESULTS is True
        assert out == ''
        assert err == ''

        with open(path, mode='r') as results_file:
            res = json.load(results_file)
        assert res == data

    def test_with_env_set_but_no_file_exists(self, monkeypatch, tmpdir, data):
        """Check that a Warning will be raised at import time,
        if environmental is set but does not correspond to an existing file.
        """
        path = str(tmpdir.join('naedw.txt'))
        monkeypatch.setenv('ORION_RESULTS_PATH', path)

        with pytest.raises(RuntimeWarning) as exc:
            reload(cli)

        assert "existing file" in str(exc.value)

    def test_call_interface_twice(self, monkeypatch, data):
        """Check that a Warning will be raised at call time,
        if function has already been called once.
        """
        monkeypatch.delenv('ORION_RESULTS_PATH', raising=False)
        reloaded_client = reload(cli)

        reloaded_client.report_results(data)
        with pytest.raises(RuntimeWarning) as exc:
            reloaded_client.report_results(data)

        assert "already reported" in str(exc.value)
        assert reloaded_client.IS_ORION_ON is False
        assert reloaded_client.RESULTS_FILENAME is None
        assert reloaded_client._HAS_REPORTED_RESULTS is True


class TestCreateExperiment:
    """Test creation of experiment with `client.create_experiment()`"""

    @pytest.mark.usefixtures("setup_pickleddb_database")
    def test_create_experiment_no_storage(self, monkeypatch):
        """Test creation if storage is not configured"""
        name = 'oopsie_forgot_a_storage'
        host = orion.core.config.storage.database.host

        with OrionState(storage=orion.core.config.storage.to_dict()) as cfg:
            # Reset the Storage and drop instances so that get_storage() would fail.
            cfg.cleanup()
            cfg.singletons = update_singletons()

            # Make sure storage must be instantiated during `create_experiment()`
            with pytest.raises(SingletonNotInstantiatedError):
                get_storage()

            experiment = create_experiment(name=name, space={'x': 'uniform(0, 10)'})

            assert isinstance(experiment._experiment._storage, Legacy)
            assert isinstance(experiment._experiment._storage._db, PickledDB)
            assert experiment._experiment._storage._db.host == host

    def test_create_experiment_new_no_space(self):
        """Test that new experiment needs space"""
        with OrionState():
            name = 'oopsie_forgot_a_space'
            with pytest.raises(NoConfigurationError) as exc:
                create_experiment(name=name)

            assert 'Experiment {} does not exist in DB'.format(name) in str(exc.value)

    def test_create_experiment_bad_storage(self):
        """Test error message if storage is not configured properly"""
        name = 'oopsie_bad_storage'
        # Make sure there is no existing storage singleton
        update_singletons()

        with pytest.raises(NotImplementedError) as exc:
            create_experiment(name=name, storage={'type': 'legacy',
                                                  'database': {'type': 'idontexist'}})

        assert "Could not find implementation of AbstractDB, type = 'idontexist'" in str(exc.value)

    def test_create_experiment_new_default(self):
        """Test creating a new experiment with all defaults"""
        name = 'all_default'
        space = {'x': 'uniform(0, 10)'}
        with OrionState():
            experiment = create_experiment(name='all_default', space=space)

            assert experiment.name == name
            assert experiment.space.configuration == space

            assert experiment.max_trials == orion.core.config.experiment.max_trials
            assert experiment.working_dir == orion.core.config.experiment.working_dir
            assert experiment.algorithms.configuration == {'random': {'seed': None}}
            assert experiment.configuration['producer'] == {'strategy': 'MaxParallelStrategy'}

    def test_create_experiment_new_full_config(self, user_config):
        """Test creating a new experiment by specifying all attributes."""
        with OrionState():
            experiment = create_experiment(**user_config)

            exp_config = experiment.configuration

            assert exp_config['space'] == config['space']
            assert exp_config['max_trials'] == config['max_trials']
            assert exp_config['working_dir'] == config['working_dir']
            assert exp_config['algorithms'] == config['algorithms']
            assert exp_config['producer'] == config['producer']

    def test_create_experiment_hit_no_branch(self, user_config):
        """Test creating an existing experiment by specifying all identical attributes."""
        with OrionState(experiments=[config]):
            experiment = create_experiment(**user_config)

            exp_config = experiment.configuration

            assert experiment.name == config['name']
            assert experiment.version == 1
            assert exp_config['space'] == config['space']
            assert exp_config['max_trials'] == config['max_trials']
            assert exp_config['working_dir'] == config['working_dir']
            assert exp_config['algorithms'] == config['algorithms']
            assert exp_config['producer'] == config['producer']

    def test_create_experiment_hit_no_config(self):
        """Test creating an existing experiment by specifying the name only."""
        with OrionState(experiments=[config]):
            experiment = create_experiment(config['name'])

            assert experiment.name == config['name']
            assert experiment.version == 1
            assert experiment.space.configuration == config['space']
            assert experiment.algorithms.configuration == config['algorithms']
            assert experiment.max_trials == config['max_trials']
            assert experiment.working_dir == config['working_dir']
            assert experiment.producer['strategy'].configuration == config['producer']['strategy']

    def test_create_experiment_hit_branch(self):
        """Test creating a differing experiment that cause branching."""
        with OrionState(experiments=[config]):
            experiment = create_experiment(config['name'], space={'y': 'uniform(0, 10)'})

            assert experiment.name == config['name']
            assert experiment.version == 2

            assert experiment.algorithms.configuration == config['algorithms']
            assert experiment.max_trials == config['max_trials']
            assert experiment.working_dir == config['working_dir']
            assert experiment.producer['strategy'].configuration == config['producer']['strategy']

    def test_create_experiment_race_condition(self, monkeypatch):
        """Test that a single race condition is handled seemlessly

        RaceCondition during registration is already handled by `build()`, therefore we will only
        test for race conditions during version update.
        """
        with OrionState(experiments=[config]):
            parent = create_experiment(config['name'])
            child = create_experiment(config['name'], space={'y': 'uniform(0, 10)'})

            def insert_race_condition(self, query):
                is_auto_version_query = (
                    query == {'name': config['name'], 'refers.parent_id': parent.id})
                if is_auto_version_query:
                    data = [child.configuration]
                # First time the query returns no other child
                elif insert_race_condition.count < 1:
                    data = [parent.configuration]
                else:
                    data = [parent.configuration, child.configuration]

                insert_race_condition.count += int(is_auto_version_query)

                return data

            insert_race_condition.count = 0

            monkeypatch.setattr(get_storage().__class__, 'fetch_experiments',
                                insert_race_condition)

            experiment = create_experiment(config['name'], space={'y': 'uniform(0, 10)'})

            assert insert_race_condition.count == 1
            assert experiment.version == 2
            assert experiment.configuration == child.configuration

    def test_create_experiment_race_condition_broken(self, monkeypatch):
        """Test that two or more race condition leads to raise"""
        with OrionState(experiments=[config]):
            parent = create_experiment(config['name'])
            child = create_experiment(config['name'], space={'y': 'uniform(0, 10)'})

            def insert_race_condition(self, query):
                is_auto_version_query = (
                    query == {'name': config['name'], 'refers.parent_id': parent.id})
                if is_auto_version_query:
                    data = [child.configuration]
                # The query returns no other child, never!
                else:
                    data = [parent.configuration]

                insert_race_condition.count += int(is_auto_version_query)

                return data

            insert_race_condition.count = 0

            monkeypatch.setattr(get_storage().__class__, 'fetch_experiments',
                                insert_race_condition)

            with pytest.raises(RaceCondition) as exc:
                create_experiment(config['name'], space={'y': 'uniform(0, 10)'})

            assert insert_race_condition.count == 2
            assert 'There was a race condition during branching and new version' in str(exc.value)

    def test_create_experiment_hit_manual_branch(self):
        """Test creating a differing experiment that cause branching."""
        new_space = {'y': 'uniform(0, 10)'}
        with OrionState(experiments=[config]):
            create_experiment(config['name'], space=new_space)

            with pytest.raises(BranchingEvent) as exc:
                create_experiment(config['name'], version=1, space=new_space)

            assert "Configuration is different and generates" in str(exc.value)

    def test_create_experiment_debug_mode(self):
        """Test that EphemeralDB is used in debug mode whatever the storage config given"""
        update_singletons()

        create_experiment(
            config['name'], space={'x': 'uniform(0, 10)'},
            storage={'type': 'legacy', 'database': {'type': 'pickleddb'}})

        storage = get_storage()

        assert isinstance(storage, Legacy)
        assert isinstance(storage._db, PickledDB)

        update_singletons()

        create_experiment(
            config['name'], space={'x': 'uniform(0, 10)'},
            storage={'type': 'legacy', 'database': {'type': 'pickleddb'}}, debug=True)

        storage = get_storage()

        assert isinstance(storage, Legacy)
        assert isinstance(storage._db, EphemeralDB)


class TestWorkon:
    """Test the helper function for sequential API"""

    def test_workon(self):
        """Verify that workon processes properly"""
        def foo(x):
            return [dict(name='result', type='objective', value=x * 2)]

        experiment = workon(foo, space={'x': 'uniform(0, 10)'}, max_trials=5)
        assert len(experiment.fetch_trials()) == 5
        assert experiment.name == 'loop'
        assert isinstance(experiment._experiment._storage, Legacy)
        assert isinstance(experiment._experiment._storage._db, EphemeralDB)

    def test_workon_algo(self):
        """Verify that algo config is processed properly"""
        def foo(x):
            return [dict(name='result', type='objective', value=x * 2)]

        experiment = workon(
            foo, space={'x': 'uniform(0, 10)'}, max_trials=5,
            algorithms={'random': {'seed': 5}})

        assert experiment.algorithms.algorithm.seed == 5

    def test_workon_name(self):
        """Verify setting the name with workon"""
        def foo(x):
            return [dict(name='result', type='objective', value=x * 2)]

        experiment = workon(foo, space={'x': 'uniform(0, 10)'}, max_trials=5, name='voici')

        assert experiment.name == 'voici'

    def test_workon_fail(self, monkeypatch):
        """Verify that storage is reverted if workon fails"""
        def foo(x):
            return [dict(name='result', type='objective', value=x * 2)]

        def build_fail(*args, **kwargs):
            raise RuntimeError('You shall not build!')

        monkeypatch.setattr('orion.core.io.experiment_builder.build', build_fail)

        # Flush storage singleton
        update_singletons()

        with pytest.raises(RuntimeError) as exc:
            experiment = workon(foo, space={'x': 'uniform(0, 10)'}, max_trials=5, name='voici')

        assert exc.match('You shall not build!')

        # Verify that tmp storage was cleared
        with pytest.raises(SingletonNotInstantiatedError):
            get_storage()

        # Now test with a prior storage
        with OrionState(storage={'type': 'legacy', 'database': {'type': 'EphemeralDB'}}):
            storage = get_storage()

            with pytest.raises(RuntimeError) as exc:
                workon(foo, space={'x': 'uniform(0, 10)'}, max_trials=5, name='voici')

            assert exc.match('You shall not build!')

            assert get_storage() is storage

    def test_workon_twice(self):
        """Verify setting the each experiment has its own storage"""
        def foo(x):
            return [dict(name='result', type='objective', value=x * 2)]

        experiment = workon(foo, space={'x': 'uniform(0, 10)'}, max_trials=5, name='voici')

        assert experiment.name == 'voici'
        assert len(experiment.fetch_trials()) == 5

        experiment2 = workon(foo, space={'x': 'uniform(0, 10)'}, max_trials=1, name='voici')

        assert experiment2.name == 'voici'
        assert len(experiment2.fetch_trials()) == 1


class TestGetExperiment:
    """Test :meth:`orion.client.get_experiment`"""

    @pytest.mark.usefixtures('mock_database')
    def test_experiment_do_not_exist(self):
        """Tests that an error is returned when the experiment doesn't exist"""
        with pytest.raises(NoConfigurationError) as exception:
            get_experiment('a')
        assert "No experiment with given name 'a' and version '*' inside database, " \
               "no view can be created." == str(exception.value)

    @pytest.mark.usefixtures('mock_database')
    def test_experiment_exist(self):
        """
        Tests that an instance of :class:`orion.core.worker.experiment.ExperimentView` is
        returned representing the latest version when none is given.
        """
        experiment = create_experiment('a', space={'x': 'uniform(0, 10)'})

        experiment = get_experiment('a')

        assert experiment
        assert isinstance(experiment, ExperimentView)

    @pytest.mark.usefixtures('mock_database')
    def test_version_do_not_exist(self, caplog):
        """Tests that a warning is printed when the experiment exist but the version doesn't"""
        create_experiment('a', space={'x': 'uniform(0, 10)'})

        experiment = get_experiment('a', 2)

        assert experiment.version == 1
        assert "Version 2 was specified but most recent version is only 1. Using 1." in caplog.text



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.io.database.ephemeraldb`."""

from datetime import datetime

import pytest

from orion.core.io.database import Database, DatabaseError, DuplicateKeyError
from orion.core.io.database.ephemeraldb import EphemeralCollection, EphemeralDB, EphemeralDocument
import orion.core.utils.backward as backward


@pytest.fixture()
def document():
    """Return EphemeralDocument."""
    return EphemeralDocument({'_id': 1, 'hello': 'there', 'mighty': 'duck'})


@pytest.fixture()
def subdocument():
    """Return EphemeralDocument with a subdocument."""
    return EphemeralDocument({'_id': 1, 'hello': 'there', 'mighty': 'duck',
                              'and': {'the': 'drake'}})


@pytest.fixture()
def collection(document):
    """Return EphemeralCollection."""
    collection = EphemeralCollection()
    collection.insert_many([document.to_dict()])

    return collection


@pytest.fixture()
def orion_db():
    """Return EphemeralDB wrapper instance initiated with test opts."""
    EphemeralDB.instance = None
    orion_db = EphemeralDB()
    return orion_db


@pytest.fixture()
def database(orion_db):
    """Return the ephemeral database, a defaultdict of ephemeral collections"""
    return orion_db._db


@pytest.fixture()
def clean_db(database, exp_config):
    """Clean insert example experiment entries to collections."""
    database['experiments'].drop()
    database['experiments'].insert_many(exp_config[0])
    database['trials'].drop()
    database['trials'].insert_many(exp_config[1])
    database['workers'].drop()
    database['workers'].insert_many(exp_config[2])
    database['resources'].drop()
    database['resources'].insert_many(exp_config[3])


@pytest.mark.usefixtures("clean_db")
class TestEnsureIndex(object):
    """Calls to :meth:`orion.core.io.database.ephemeraldb.EphemeralDB.ensure_index`."""

    def test_new_index(self, orion_db):
        """Index should be added to ephemeral database"""
        assert "new_field_1" not in orion_db._db['new_collection']._indexes

        orion_db.ensure_index('new_collection', 'new_field', unique=False)
        assert "new_field_1" not in orion_db._db['new_collection']._indexes

        orion_db.ensure_index('new_collection', 'new_field', unique=True)
        assert "new_field_1" in orion_db._db['new_collection']._indexes

    def test_existing_index(self, orion_db):
        """Index should be added to ephemeral database and reattempt should do nothing"""
        assert "new_field_1" not in orion_db._db['new_collection']._indexes

        orion_db.ensure_index('new_collection', 'new_field', unique=True)
        assert "new_field_1" in orion_db._db['new_collection']._indexes

        # reattempt
        orion_db.ensure_index('new_collection', 'new_field', unique=True)
        assert "new_field_1" in orion_db._db['new_collection']._indexes

    def test_compound_index(self, orion_db):
        """Tuple of Index should be added as a compound index."""
        assert "name_1_metadata.user_1" not in orion_db._db['experiments']._indexes
        orion_db.ensure_index('experiments',
                              [('name', Database.ASCENDING),
                               ('metadata.user', Database.ASCENDING)], unique=True)
        assert "name_1_metadata.user_1" in orion_db._db['experiments']._indexes


@pytest.mark.usefixtures("clean_db")
class TestRead(object):
    """Calls to :meth:`orion.core.io.database.ephemeraldb.EphemeralDB.read`."""

    def test_read_experiment(self, exp_config, orion_db):
        """Fetch a whole experiment entries."""
        loaded_config = orion_db.read(
            'trials', {'experiment': 'supernaedo2-dendi', 'status': 'new'})
        assert loaded_config == [exp_config[1][3], exp_config[1][4]]

        loaded_config = orion_db.read(
            'trials',
            {'experiment': 'supernaedo2-dendi',
             'submit_time': exp_config[1][3]['submit_time']})
        assert loaded_config == [exp_config[1][3]]
        assert loaded_config[0]['_id'] == exp_config[1][3]['_id']

    def test_read_with_id(self, exp_config, orion_db):
        """Query using ``_id`` key."""
        loaded_config = orion_db.read('experiments', {'_id': exp_config[0][2]['_id']})
        backward.populate_space(loaded_config[0])
        assert loaded_config == [exp_config[0][2]]

    def test_read_default(self, exp_config, orion_db):
        """Fetch value(s) from an entry."""
        value = orion_db.read(
            'experiments', {'name': 'supernaedo2', 'metadata.user': 'tsirif'},
            selection={'algorithms': 1, '_id': 0})
        assert value == [{'algorithms': exp_config[0][0]['algorithms']}]

    def test_read_nothing(self, orion_db):
        """Fetch value(s) from an entry."""
        value = orion_db.read(
            'experiments', {'name': 'not_found', 'metadata.user': 'tsirif'},
            selection={'algorithms': 1})
        assert value == []

    def test_read_trials(self, exp_config, orion_db):
        """Fetch value(s) from an entry."""
        value = orion_db.read(
            'trials',
            {'experiment': 'supernaedo2-dendi',
             'submit_time': {'$gte': datetime(2017, 11, 23, 0, 0, 0)}})
        assert value == [exp_config[1][1]] + exp_config[1][3:7]

        value = orion_db.read(
            'trials',
            {'experiment': 'supernaedo2-dendi',
             'submit_time': {'$gt': datetime(2017, 11, 23, 0, 0, 0)}})
        assert value == exp_config[1][3:7]

    def test_null_comp(self, exp_config, orion_db):
        """Fetch value(s) from an entry."""
        all_values = orion_db.read(
            'trials',
            {'experiment': 'supernaedo2-dendi',
             'end_time': {'$gte': datetime(2017, 11, 1, 0, 0, 0)}})

        orion_db._db['trials']._documents[0]._data['end_time'] = None
        values = orion_db.read(
            'trials',
            {'experiment': 'supernaedo2-dendi',
             'end_time': {'$gte': datetime(2017, 11, 1, 0, 0, 0)}})
        assert len(values) == len(all_values) - 1


@pytest.mark.usefixtures("clean_db")
class TestWrite(object):
    """Calls to :meth:`orion.core.io.database.ephemeraldb.EphemeralDB.write`."""

    def test_insert_one(self, database, orion_db):
        """Should insert a single new entry in the collection."""
        item = {'exp_name': 'supernaekei',
                'user': 'tsirif'}
        count_before = database['experiments'].count()
        # call interface
        assert orion_db.write('experiments', item) == 1
        assert database['experiments'].count() == count_before + 1
        value = database['experiments'].find({'exp_name': 'supernaekei'})[0]
        assert value == item

    def test_insert_many(self, database, orion_db):
        """Should insert two new entry (as a list) in the collection."""
        item = [{'exp_name': 'supernaekei2',
                 'user': 'tsirif'},
                {'exp_name': 'supernaekei3',
                 'user': 'tsirif'}]
        count_before = database['experiments'].count()
        # call interface
        assert orion_db.write('experiments', item) == 2
        assert database['experiments'].count() == count_before + 2
        value = database['experiments'].find({'exp_name': 'supernaekei2'})[0]
        assert value == item[0]
        value = database['experiments'].find({'exp_name': 'supernaekei3'})[0]
        assert value == item[1]

    def test_update_many_default(self, database, orion_db):
        """Should match existing entries, and update some of their keys."""
        filt = {'metadata.user': 'dendi'}
        count_before = database['experiments'].count()
        count_query = database['experiments'].count(filt)
        # call interface
        assert orion_db.write('experiments', {'pool_size': 16}, filt) == count_query
        assert database['experiments'].count() == count_before
        value = list(database['experiments'].find({}))
        assert value[0]['pool_size'] == 16
        assert value[1]['pool_size'] == 2
        assert value[2]['pool_size'] == 2
        assert value[3]['pool_size'] == 2

    def test_update_with_id(self, exp_config, database, orion_db):
        """Query using ``_id`` key."""
        filt = {'_id': exp_config[0][1]['_id']}
        count_before = database['experiments'].count()
        # call interface
        assert orion_db.write('experiments', {'pool_size': 36}, filt) == 1
        assert database['experiments'].count() == count_before
        value = list(database['experiments'].find())
        assert value[0]['pool_size'] == 2
        assert value[1]['pool_size'] == 36
        assert value[2]['pool_size'] == 2

    def test_insert_duplicate(self, database, orion_db):
        """Verify that duplicates cannot by inserted if index is unique"""
        orion_db.ensure_index('some_doc', 'unique_field', unique=True)
        # call interface
        orion_db.write('some_doc', {'unique_field': 1})

        with pytest.raises(DuplicateKeyError) as exc:
            orion_db.write('some_doc', {'unique_field': 1})

        assert 'unique_field' in str(exc.value)


@pytest.mark.usefixtures("clean_db")
class TestReadAndWrite(object):
    """Calls to :meth:`orion.core.io.database.ephemeraldb.EphemeralDB.read_and_write`."""

    def test_read_and_write_one(self, database, orion_db, exp_config):
        """Should read and update a single entry in the collection."""
        # Make sure there is only one match
        documents = orion_db.read(
            'experiments',
            {'name': 'supernaedo4'})
        assert len(documents) == 1

        # Find and update atomically
        loaded_config = orion_db.read_and_write(
            'experiments',
            {'name': 'supernaedo4'},
            {'pool_size': 'lalala'})
        exp_config[0][3]['pool_size'] = 'lalala'
        backward.populate_space(loaded_config)
        assert loaded_config == exp_config[0][3]

    def test_read_and_write_many(self, database, orion_db, exp_config):
        """Should update only one entry."""
        # Make sure there is many matches
        documents = orion_db.read('experiments', {'metadata.user': 'tsirif'})
        assert len(documents) > 1

        # Find many and update first one only
        loaded_config = orion_db.read_and_write(
            'experiments',
            {'metadata.user': 'tsirif'},
            {'pool_size': 'lalala'})

        exp_config[0][1]['pool_size'] = 'lalala'
        backward.populate_space(loaded_config)
        assert loaded_config == exp_config[0][1]

        # Make sure it only changed the first document found
        documents = orion_db.read('experiments', {'metadata.user': 'tsirif'})
        assert documents[0]['pool_size'] == 'lalala'
        assert documents[1]['pool_size'] != 'lalala'

    def test_read_and_write_no_match(self, database, orion_db):
        """Should return None when there is no match."""
        loaded_config = orion_db.read_and_write(
            'experiments',
            {'name': 'lalala'},
            {'pool_size': 'lalala'})

        assert loaded_config is None


@pytest.mark.usefixtures("clean_db")
class TestRemove(object):
    """Calls to :meth:`orion.core.io.database.ephemeraldb.EphemeralDB.remove`."""

    def test_remove_many_default(self, exp_config, database, orion_db):
        """Should match existing entries, and delete them all."""
        filt = {'metadata.user': 'tsirif'}
        count_before = database['experiments'].count()
        count_filt = database['experiments'].count(filt)
        # call interface
        assert orion_db.remove('experiments', filt) == count_filt
        assert database['experiments'].count() == count_before - count_filt
        assert database['experiments'].count() == 1
        loaded_config = list(database['experiments'].find())
        backward.populate_space(loaded_config[0])
        assert loaded_config == [exp_config[0][0]]

    def test_remove_with_id(self, exp_config, database, orion_db):
        """Query using ``_id`` key."""
        filt = {'_id': exp_config[0][0]['_id']}

        count_before = database['experiments'].count()
        # call interface
        assert orion_db.remove('experiments', filt) == 1
        assert database['experiments'].count() == count_before - 1
        loaded_configs = database['experiments'].find()
        for loaded_config in loaded_configs:
            backward.populate_space(loaded_config)
        assert loaded_configs == exp_config[0][1:]

    def test_remove_update_indexes(self, exp_config, database, orion_db):
        """Verify that indexes are properly update after deletion."""
        with pytest.raises(DuplicateKeyError):
            orion_db.write('experiments', {'_id': exp_config[0][0]['_id']})
        with pytest.raises(DuplicateKeyError):
            orion_db.write('experiments', {'_id': exp_config[0][1]['_id']})

        filt = {'_id': exp_config[0][0]['_id']}

        count_before = database['experiments'].count()
        # call interface
        assert orion_db.remove('experiments', filt) == 1
        assert database['experiments'].count() == count_before - 1
        # Should not fail now, otherwise it means the indexes were not updated properly during
        # remove()
        orion_db.write('experiments', filt)
        # And this should still fail
        with pytest.raises(DuplicateKeyError):
            orion_db.write('experiments', {'_id': exp_config[0][1]['_id']})


@pytest.mark.usefixtures("clean_db")
class TestCount(object):
    """Calls :meth:`orion.core.io.database.ephemeraldb.EphemeralDB.count`."""

    def test_count_default(self, exp_config, orion_db):
        """Call just with collection name."""
        found = orion_db.count('trials')
        assert found == len(exp_config[1])

    def test_count_query(self, exp_config, orion_db):
        """Call with a query."""
        found = orion_db.count('trials', {'status': 'completed'})
        assert found == len([x for x in exp_config[1] if x['status'] == 'completed'])

    def test_count_query_with_id(self, exp_config, orion_db):
        """Call querying with unique _id."""
        found = orion_db.count('trials', {'_id': exp_config[1][2]['_id']})
        assert found == 1

    def test_count_nothing(self, orion_db):
        """Call with argument that will not find anything."""
        found = orion_db.count('experiments', {'name': 'lalalanotfound'})
        assert found == 0


@pytest.mark.usefixtures("clean_db")
class TestIndex(object):
    """Test index for :meth:`orion.core.io.database.ephemeraldb.EphemeralCollection`."""

    def test_create_index(self, collection):
        """Test if new index added property."""
        collection.create_index('hello')
        assert collection._indexes == {'_id_': (('_id',), {(1, )})}

        collection.create_index('hello', unique=True)
        assert collection._indexes == {'_id_': (('_id',), {(1, )}),
                                       'hello_1': (('hello', ), {('there', )})}

    def test_track_index(self, collection):
        """Test if index values are tracked property."""
        collection.create_index('hello', unique=True)
        collection.insert_many([{'hello': 'here'}, {'hello': 2}])
        assert (
            collection._indexes ==
            {'_id_': (('_id',), {(1, ), (2, ), (3, )}),
             'hello_1': (('hello', ), {('there', ), ('here', ), (2, )})})

    def test_index_over_non_existing_field(self, collection):
        """Test if index values are tracked property."""
        collection.create_index([('hello', EphemeralDB.DESCENDING),
                                 ('idontexist', EphemeralDB.ASCENDING)], unique=True)

        collection.insert_many([{'hello': 'here'}, {'hello': 2}])
        assert (
            collection._indexes ==
            {'_id_': (('_id',), {(1, ), (2, ), (3, )}),
             'hello_1_idontexist_1': (('hello', 'idontexist'),
                                      {('there', None), ('here', None), (2, None)})})
        assert (
            collection.find({}, selection={'hello': 1, 'idontexist': 1}) ==
            [{'_id': 1, 'hello': 'there', 'idontexist': None},
             {'_id': 2, 'hello': 'here', 'idontexist': None},
             {'_id': 3, 'hello': 2, 'idontexist': None}])


@pytest.mark.usefixtures("clean_db")
class TestSelect(object):
    """Calls :meth:`orion.core.io.database.ephemeraldb.EphemeralDocument.select`."""

    def test_select_all(self, document):
        """Select only one field."""
        assert document.select({}) == {'_id': 1, 'hello': 'there', 'mighty': 'duck'}

    def test_select_id(self, document):
        """Select only one field."""
        assert document.select({'_id': 1}) == {'_id': 1}

    def test_select_one(self, document):
        """Select only one field."""
        assert document.select({'hello': 1}) == {'_id': 1, 'hello': 'there'}

    def test_select_two(self, document):
        """Select only two field."""
        assert (
            document.select({'hello': 1, 'mighty': 1}) ==
            {'_id': 1, 'hello': 'there', 'mighty': 'duck'})

    def test_unselect_one(self, document):
        """Unselect only one field."""
        assert document.select({'hello': 0}) == {'_id': 1, 'mighty': 'duck'}

    def test_unselect_two(self, document):
        """Unselect two field."""
        assert document.select({'_id': 0, 'hello': 0}) == {'mighty': 'duck'}

    def test_mixed_select(self, document):
        """Select one field and unselect _id."""
        assert document.select({'_id': 0, 'hello': 1}) == {'hello': 'there'}

    def test_select_unexisting_field(self, document):
        """Select field that does not exist and should return None."""
        assert document.select({'idontexist': 1}) == {'_id': 1, 'idontexist': None}


@pytest.mark.usefixtures('clean_db')
class TestMatch:
    """Calls :meth:`orion.core.io.database.ephemeraldb.EphemeralDocument.match`."""

    def test_match_eq(self, document):
        """Test eq operator"""
        assert document.match({'hello': 'there'})
        assert not document.match({'hello': 'not there'})

    def test_match_sub_eq(self, subdocument):
        """Test eq operator with sub document"""
        assert subdocument.match({'and.the': 'drake'})
        assert not subdocument.match({'and.no': 'drake'})

    def test_match_in(self, subdocument):
        """Test $in operator with document"""
        assert subdocument.match({'hello': {'$in': ['there', 'here']}})
        assert not subdocument.match({'hello': {'$in': ['ici', 'here']}})

    def test_match_sub_in(self, subdocument):
        """Test $in operator with sub document"""
        assert subdocument.match({'and.the': {'$in': ['duck', 'drake']}})
        assert not subdocument.match({'and.the': {'$in': ['hyppo', 'lion']}})

    def test_match_gte(self, document):
        """Test $gte operator with document"""
        assert document.match({'_id': {'$gte': 1}})
        assert document.match({'_id': {'$gte': 0}})
        assert not document.match({'_id': {'$gte': 2}})

    def test_match_gt(self, document):
        """Test $gt operator with document"""
        assert document.match({'_id': {'$gt': 0}})
        assert not document.match({'_id': {'$gt': 1}})

    def test_match_lte(self, document):
        """Test $lte operator with document"""
        assert document.match({'_id': {'$lte': 2}})
        assert document.match({'_id': {'$lte': 1}})
        assert not document.match({'_id': {'$lte': 0}})

    def test_match_ne(self, document):
        """Test $ne operator with document"""
        assert document.match({'hello': {'$ne': 'here'}})
        assert not document.match({'hello': {'$ne': 'there'}})

    def test_match_bad_operator(self, document):
        """Test invalid operator handling"""
        with pytest.raises(ValueError) as exc:
            document.match({'_id': {'$voici_voila': 0}})

        assert 'Operator \'$voici_voila\' is not supported' in str(exc.value)


@pytest.mark.usefixtures("clean_db")
class TestIndexInformation(object):
    """Calls :meth:`orion.core.io.database.mongodb.EphemeralDB.count`."""

    def test_no_index(self, orion_db):
        """Test that no index is returned when there is none."""
        assert orion_db.index_information('experiments') == {'_id_': True}

    def test_single_index(self, orion_db):
        """Test that single indexes are ignored if not unique."""
        orion_db.ensure_index('experiments', [('name', EphemeralDB.ASCENDING)])

        assert orion_db.index_information('experiments') == {'_id_': True}

    def test_single_index_unique(self, orion_db):
        """Test with single unique indexes."""
        orion_db.ensure_index('experiments', [('name', EphemeralDB.ASCENDING)], unique=True)

        assert orion_db.index_information('experiments') == {'_id_': True, 'name_1': True}

    def test_ordered_index(self, orion_db):
        """Test that ordered indexes are not taken into account."""
        orion_db.ensure_index('experiments', [('name', EphemeralDB.DESCENDING)], unique=True)

        assert orion_db.index_information('experiments') == {'_id_': True, 'name_1': True}

    def test_compound_index(self, orion_db):
        """Test representation of compound indexes."""
        orion_db.ensure_index('experiments',
                              [('name', EphemeralDB.DESCENDING),
                               ('version', EphemeralDB.ASCENDING)], unique=True)

        index_info = orion_db.index_information('experiments')
        assert index_info == {'_id_': True, 'name_1_version_1': True}


@pytest.mark.usefixtures("clean_db")
class TestDropIndex(object):
    """Calls :meth:`orion.core.io.database.mongodb.EphemeralDB.count`."""

    def test_no_index(self, orion_db):
        """Test that no index is returned when there is none."""
        with pytest.raises(DatabaseError) as exc:
            orion_db.drop_index('experiments', 'i_dont_exist')
        assert 'index not found with name' in str(exc.value)

    def test_drop_single_index(self, orion_db):
        """Test with single indexes."""
        orion_db.ensure_index('experiments', [('name', EphemeralDB.ASCENDING)], unique=True)
        assert orion_db.index_information('experiments') == {'_id_': True, 'name_1': True}
        orion_db.drop_index('experiments', 'name_1')
        assert orion_db.index_information('experiments') == {'_id_': True}

    def test_drop_ordered_single_index(self, orion_db):
        """Test with single indexes."""
        orion_db.ensure_index('experiments', [('name', EphemeralDB.DESCENDING)], unique=True)
        index_info = orion_db.index_information('experiments')
        assert index_info == {'_id_': True, 'name_1': True}
        orion_db.drop_index('experiments', 'name_1')
        index_info = orion_db.index_information('experiments')
        assert index_info == {'_id_': True}

    def test_drop_ordered_compound_index(self, orion_db):
        """Test with single indexes."""
        orion_db.ensure_index('experiments',
                              [('name', EphemeralDB.ASCENDING),
                               ('version', EphemeralDB.DESCENDING)],
                              unique=True)
        index_info = orion_db.index_information('experiments')
        assert index_info == {'_id_': True, 'name_1_version_1': True}
        orion_db.drop_index('experiments', 'name_1_version_1')
        index_info = orion_db.index_information('experiments')
        assert index_info == {'_id_': True}



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.worker.trial_monitor` -- Monitor trial execution
=================================================================
.. module:: trial_monitor
   :platform: Unix
   :synopsis: Monitor trials and update their heartbeat

"""
import threading

from orion.storage.base import get_storage

STOPPED_STATUS = {'completed', 'interrupted', 'suspended'}


class TrialPacemaker(threading.Thread):
    """Monitor a given trial inside a thread, updating its heartbeat
    at a given interval of time.

    Parameters
    ----------
    exp: Experiment
        The current Experiment.

    """

    def __init__(self, trial, wait_time=60):
        threading.Thread.__init__(self)
        self.stopped = threading.Event()
        self.trial = trial
        self.wait_time = wait_time
        self.storage = get_storage()

    def stop(self):
        """Stop monitoring."""
        self.stopped.set()
        self.join()

    def run(self):
        """Run the trial monitoring every given interval."""
        while not self.stopped.wait(self.wait_time):
            self._monitor_trial()

    def _monitor_trial(self):
        trial = self.storage.get_trial(self.trial)

        if trial.status in STOPPED_STATUS:
            self.stopped.set()
        else:
            if not self.storage.update_heartbeat(trial):
                self.stopped.set()
